#Standard python libraries
import os
import warnings
import copy
import time
import itertools
import functools

#Dependencies - numpy, scipy, matplotlib, pyfftw
import numpy as np
import matplotlib.pyplot as plt
try:
    from pyfftw.interfaces.numpy_fft import fft, fftshift, ifft, ifftshift, fftfreq
except ImportError:
    from numpy.fft import fft, fftshift, ifft, ifftshift, fftfreq
from scipy.interpolate import interp1d as sinterp1d
import scipy
from scipy.sparse import save_npz, load_npz, eye, csr_matrix
from scipy.sparse.linalg import eigs

from ufss import DiagramGenerator
from scipy.integrate import RK45

class RK_rho_container:
    def __init__(self,t,rho,pulse_number,manifold_key,*,interp_kind='linear',
                 optical_gap = 0):
        self.pulse_number = pulse_number
        self.n, self.M = rho.shape
        self.manifold_key = manifold_key
        self.optical_gap = optical_gap
        if t.size == 1:
            self.M = self.M+2
            self.t = np.array([-1,0,1],dtype='float') * np.spacing(t[0]) + t[0]
            rho_new = np.zeros((self.n,3),dtype='complex')
            rho_new[:,0] = 0
            rho_new[:,1] = 0.5 * rho[:,0]
            rho_new[:,2] = rho[:,0]
            self.rho = rho_new
            
            self.interp = self.make_interpolant(kind='zero')
            
        else:
            self.t = t
            self.rho = rho

            self.interp = self.make_interpolant(kind=interp_kind)

        self.t_checkpoint = t
        self.rho_checkpoint = rho

    def make_interpolant(self,*, kind='cubic'):
        """Interpolates density matrix
"""
        return sinterp1d(self.t,self.rho,fill_value = (0,np.nan),bounds_error = False,
                         assume_sorted=True,kind=kind)

    def one_time_step(self,rho0,t0,tf,*,find_best_starting_time = True):
        if find_best_starting_time and tf < self.t_checkpoint[-1]:
            diff1 = tf - t0

            diff2 = tf - self.t[-1]

            closest_t_checkpoint_ind = np.argmin(np.abs(self.t_checkpoint - tf))
            closest_t_checkpoint = self.t_checkpoint[closest_t_checkpoint_ind]
            diff3 = tf - closest_t_checkpoint

            rho0s = [rho0,self.rho[:,-1],self.rho_checkpoint[:,closest_t_checkpoint_ind]]
            
            neighbor_ind = closest_t_checkpoint_ind - 1
            if neighbor_ind >= 0:
                neighbor = self.t_checkpoint[closest_t_checkpoint_ind-1]
                diff4 = tf - neighbor
                rho0s.append(self.rho_checkpoint[:,neighbor_ind])
            else:
                neighbor = np.nan
                diff4 = np.inf
                

            t0s = np.array([t0,self.t[-1],closest_t_checkpoint,neighbor])
            diffs = np.array([diff1,diff2,diff3,diff4])
            
            for i in range(diffs.size):
                if diffs[i] < 0:
                    diffs[i] = np.inf
            
            if np.allclose(diffs,np.inf):
                raise ValueError('Method extend is only valid for times after the pulse has ended')
            
            t0 = t0s[np.argmin(diffs)]
            rho0 = rho0s[np.argmin(diffs)]
            
        elif find_best_starting_time and tf > self.t_checkpoint[-1]:
            if self.t_checkpoint[-1] > t0:
                t0 = self.t_checkpoint[-1]
                rho0 = self.rho_checkpoint[:,-1]
            else:
                pass
            
        else:
            pass
        # RWA_gap = self.manifold.dot(np.array([1,-1])) * self.optical_gap
        return self.one_time_step_function(rho0,t0,tf,manifold_key=self.manifold_key)#,RWA_gap = RWA_gap)

    def extend(self,t):
        ans = np.zeros((self.n,t.size),dtype='complex')
        
        if t[0] >= self.t_checkpoint[0]:

            t_intersect, t_inds, t_checkpoint_inds = np.intersect1d(t,self.t_checkpoint,return_indices=True)

            ans[:,t_inds] = self.rho_checkpoint[:,t_checkpoint_inds]

            if t_inds.size == t.size:
                return ans
            else:
                all_t_inds = np.arange(t.size)
                other_t_inds = np.setdiff1d(all_t_inds,t_inds)
                t0 = self.t_checkpoint[-1]
                rho0 = self.rho_checkpoint[:,-1]
                if t[other_t_inds[0]] >= t0:
                    find_best_starting_time = False
                else:
                    find_best_starting_time = True
                for t_ind in other_t_inds:
                    tf = t[t_ind]
                    ans[:,t_ind] = self.one_time_step(rho0,t0,tf,find_best_starting_time = find_best_starting_time)
                    t0 = tf
                    rho0 = ans[:,t_ind]
            
        elif t[0] >= self.t[-1]:
            t0 = self.t[-1]
            rho0 = self.rho[:,-1]
            for i in range(len(t)):
                ans[:,i] = self.one_time_step(rho0,t0,t[i],find_best_starting_time = True)
                t0 = t[i]
                rho0 = ans[:,i]
        else:
            raise ValueError('Method extend is only valid for times after the pulse has ended')

        self.rho_checkpoint = ans
        self.t_checkpoint = t
        return ans

    def __call__(self,t):
        """Assumes t is sorted """
        if type(t) is np.ndarray:
            pass
        elif type(t) is list:
            t = np.array(t)
        else:
            t = np.array([t])
        extend_inds = np.where(t>self.t[-1])
        interp_inds = np.where(t<=self.t[-1])
        ta = t[interp_inds]
        tb = t[extend_inds]
        if ta.size > 0:
            ans_a_flag = True
            if ta.size == self.M and np.allclose(ta,self.t):
                ans_a = self.rho
            else:
                ans_a = self.interp(ta)
        else:
            ans_a_flag = False
        if tb.size > 0:
            ans_b = self.extend(tb)
            ans_b_flag = True
        else:
            ans_b_flag = False
            
        if ans_a_flag and ans_b_flag:
            ans = np.hstack((ans_a,ans_b))
        elif ans_a_flag:
            ans = ans_a
        elif ans_b_flag:
            ans = ans_b
        else:
            ans = None
        return ans

    def __getitem__(self,inds):
        return self.rho[:,inds]

class RKE_DensityMatrices(DiagramGenerator):
    """This class is designed to calculate perturbative wavepackets in the
        light-matter interaction given the eigenvalues of the unperturbed 
        hamiltonian and the material dipole operator evaluated in the
        eigenbasis of the unperturbed hamiltonian.

    Args:
        file_path (string): path to folder containing eigenvalues and the
            dipole operator for the system Hamiltonian
        detection_type (string): options are 'polarization' (default) or 'fluorescence'

"""
    def __init__(self,file_path,*,detection_type = 'polarization',
                 conserve_memory=False):
        self.slicing_time = 0
        self.interpolation_time = 0
        self.expectation_time = 0
        self.RK45_step_time = 0
        self.dipole_dot_rho_time = 0
        self.dipole_time = 0
        self.automation_time = 0
        self.diagram_to_signal_time = 0
        
        self.base_path = file_path

        self.undersample_factor = 1

        self.gamma_res = 6.91

        self.sparsity_threshold = 0.1

        self.conserve_memory = conserve_memory

        self.load_L()

        self.set_rho_shapes()

        self.load_H_eigensystem()

        if not self.conserve_memory:
            self.load_mu()

        try:
            self.load_H_mu()
            # more efficient if H_mu is available
            self.dipole_down = self.dipole_down_H_mu

        except:
            # generally less efficient - mostly here for backwards compatibility
            self.dipole_down = self.dipole_down_L_mu

        self.optical_gap = 0

        self.atol = 1E-6
        self.rtol = 1E-5

        if detection_type == 'polarization':
            self.rho_to_signal = self.polarization_detection_rho_to_signal
            self.return_complex_signal = False
            
        elif detection_type == 'complex_polarization':
            self.rho_to_signal = self.polarization_detection_rho_to_signal
            self.return_complex_signal = True
            detection_type = 'polarization'
            
        elif detection_type == 'integrated_polarization':
            raise Exception('detection_type: Integrated polarization is not implemented for Open RKE')
            self.rho_to_signal = self.integrated_polarization_detection_rho_to_signal
            self.return_complex_signal = False
            
        elif detection_type == 'fluorescence':
            self.rho_to_signal = self.fluorescence_detection_rho_to_signal

        DiagramGenerator.__init__(self,detection_type=detection_type)
        self.KB_dict = {'Bu':self.bra_up,'Ku':self.ket_up,'Kd':self.ket_down,'Bd':self.bra_down}

        # Code will not actually function until the following three empty lists are set by the user
        self.efields = [] #initialize empty list of electric field shapes
        self.efield_times = [] #initialize empty list of times assoicated with each electric field shape
        self.dts = [] #initialize empty list of time spacings associated with each electric field shape
        self.polarization_sequence = [] #initialize empty polarization sequence
        self.pulse_times = [] #initialize empty list of pulse arrival times
        self.centers = [] #initialize empty list of pulse center frequencies
        self.efield_wavevectors = []
        self.rhos = dict()
        
        # Initialize unperturbed wavefunction
        self.set_rho0_auto()

    def set_pulse_delays(self,all_delays):
        """Must be a list of numpy arrays, where each array is a
            list of delay times between pulses
"""
        self.all_pulse_delays = all_delays
        num_delays = len(self.all_pulse_delays)
        num_pulses = len(self.efields)
        
        if num_delays == num_pulses - 1:
            pass
        elif num_delays == num_pulses - 2 and self.detection_type == 'polarization':
            # If there is a local oscillator, it arrives simultaneously with the last pulse
            self.all_pulse_delays.append(np.array([0]))
        elif num_delays <= num_pulses -2:
            raise Exception('There are not enough delay times')
        elif num_delays >= num_pulses:
            raise Exception('There are too many delay times')

    def calculate_diagrams_all_delays(self,diagrams):
        t0 = time.time()
        num_delays = len(self.all_pulse_delays)
        num_pulses = len(self.efields)

        all_delay_combinations = list(itertools.product(*self.all_pulse_delays))
        
        signal_shape = [delays.size for delays in self.all_pulse_delays]
        if self.detection_type == 'polarization':
            signal = np.zeros((len(all_delay_combinations),self.w.size),dtype='complex')
            if len(signal_shape) == self.pdc.shape[0]:
                # get rid of the "delay" between the last pulse and the local oscillator
                signal_shape[-1] = self.w.size
            elif len(signal_shape) == self.pdc.shape[0] - 1:
                # append the shape of the polariation-detection axis
                signal_shape.append(self.w.size)
            else:
                raise Exception('Cannot automatically determine final signal shape')
        else:
            signal = np.zeros((len(all_delay_combinations)),dtype='complex')

        counter = 0
        for delays in all_delay_combinations:
            arrival_times = [0]
            for delay in delays:
                arrival_times.append(arrival_times[-1]+delay)

            if self.detection_type == 'polarization':
                signal[counter,:] = self.calculate_diagrams(diagrams,arrival_times)
            else:
                signal[counter] = self.calculate_diagrams(diagrams,arrival_times)
            counter += 1

        self.signal = signal.reshape(signal_shape)
        self.calculation_time = time.time() - t0
        
        return self.signal

    def save_timing(self):
        save_dict = {'RKE_calculation_time':self.calculation_time}
        np.savez(os.path.join(self.base_path,'RKE_calculation_time.npz'),**save_dict)
        
    def calculate_signal_all_delays(self):
        t0 = time.time()
        num_delays = len(self.all_pulse_delays)
        num_pulses = len(self.efields)

        all_delay_combinations = list(itertools.product(*self.all_pulse_delays))
        
        signal_shape = [delays.size for delays in self.all_pulse_delays]
        if self.detection_type == 'polarization':
            signal = np.zeros((len(all_delay_combinations),self.w.size),dtype='complex')
            if len(signal_shape) == self.pdc.shape[0]:
                # get rid of the "delay" between the last pulse and the local oscillator
                signal_shape[-1] = self.w.size
            elif len(signal_shape) == self.pdc.shape[0] - 1:
                # append the shape of the polariation-detection axis
                signal_shape.append(self.w.size)
            else:
                raise Exception('Cannot automatically determine final signal shape')
        else:
            signal = np.zeros((len(all_delay_combinations)),dtype='complex')

        counter = 0
        for delays in all_delay_combinations:
            arrival_times = [0]
            for delay in delays:
                arrival_times.append(arrival_times[-1]+delay)

            if self.detection_type == 'polarization':
                signal[counter,:] = self.calculate_signal(arrival_times)
            else:
                signal[counter] = self.calculate_signal(arrival_times)
            counter += 1

        self.signal = signal.reshape(signal_shape)
        self.calculation_time = time.time() - t0
        return self.signal

    def set_t(self,optical_dephasing_rate,*,dt='auto'):
        """Sets the time grid upon which all frequency-detected signals will
be calculated on
"""
        max_pos_t = int(self.gamma_res/optical_dephasing_rate)
        if dt == 'auto':
            dt = self.dts[-1] # signal detection bandwidth determined by local oscillator
        self.t = np.arange(-max_pos_t,max_pos_t+dt/2,dt)
        # if self.t.size % 2:
        #     self.t = self.t[:-1]
        self.w = fftshift(fftfreq(self.t.size,d=dt)*2*np.pi)

    def execute_diagram(self,instructions):
        r = self.rho0
        name = ''
        for i in range(len(instructions)):
            key, num = instructions[i]
            name += key+str(num)
            # Try to re-use previous calculations, if they exist
            try:
                new_r = self.rhos[name]
            except KeyError:
                new_r = self.KB_dict[key](r,pulse_number=num)
                self.rhos[name] = new_r
            r = new_r
        sig = self.rho_to_signal(r)
        return sig

    def remove_rhos_by_pulse_number(self,pulse_number):
        num = str(pulse_number)
        keys = self.rhos.keys()
        keys_to_remove = []
        for key in keys:
            flag = key.find(num)
            if flag >= 0:
                keys_to_remove.append(key)
        for key in keys_to_remove:
            self.rhos.pop(key)

    def calculate_signal(self,arrival_times):
        t0 = time.time()
        try:
            old_pulse_times = self.pulse_times
            for i in range(len(old_pulse_times)):
                if old_pulse_times[i] != arrival_times[i]:
                    self.remove_rhos_by_pulse_number(i)
        except AttributeError:
            pass
        
        self.pulse_times = arrival_times
        if self.detection_type == 'polarization':
            times = [self.efield_times[i] + arrival_times[i] for i in range(len(arrival_times)-1)]
        elif self.detection_type == 'integrated_polarization':
            times = [self.efield_times[i] + arrival_times[i] for i in range(len(arrival_times)-1)]
        elif self.detection_type == 'fluorescence':
            times = [self.efield_times[i] + arrival_times[i] for i in range(len(arrival_times))]
        
        efield_permutations = self.relevant_permutations(times)
        
        diagram_instructions = []
        for perm in efield_permutations:
            diagram_instructions += self.instructions_from_permutation(perm)
        self.current_instructions = diagram_instructions

        t1 = time.time()
        try:
            instructions = diagram_instructions[0]
            signal = self.execute_diagram(instructions)
            for instructions in diagram_instructions[1:]:
                signal += self.execute_diagram(instructions)
        except IndexError:
            signal = 0

        t2 = time.time()
        self.automation_time += t1-t0
        self.diagram_to_signal_time += t2-t1
        return signal

    def calculate_diagrams(self,diagram_instructions,arrival_times):
        try:
            old_pulse_times = self.pulse_times
            for i in range(len(old_pulse_times)):
                if old_pulse_times[i] != arrival_times[i]:
                    self.remove_rhos_by_pulse_number(i)
        except AttributeError:
            pass
        
        self.pulse_times = arrival_times
            
        self.current_instructions = diagram_instructions
        instructions = diagram_instructions[0]
        signal = self.execute_diagram(instructions)
        for instructions in diagram_instructions[1:]:
            signal += self.execute_diagram(instructions)
        return signal

    def polarization_detection_rho_to_signal(self,rho):    
        p_of_t = self.dipole_expectation(rho,pulse_number=-1,ket_flag=True)
        return self.polarization_to_signal(p_of_t,local_oscillator_number=-1)

    def integrated_polarization_detection_rho_to_signal(self,rho):    
        p = self.integrated_dipole_expectation(rho,ket_flag=True)
        return self.integrated_polarization_to_signal(p,local_oscillator_number=-1)

    # def fluorescence_detection_rho_to_signal(self,rho):
    #     L_size = self.eigenvalues[0].size
    #     H_size = int(np.sqrt(L_size))

    #     # reshape rho into a normal density matrix representation
    #     rho = rho.reshape((H_size,H_size))

    #     fluorescence_yield = np.array([0,1,1,self.f_yield])

    #     signal = np.dot(np.diagonal(rho),fluorescence_yield)
        
    #     return signal

    def set_efields(self,times_list,efields_list,centers_list,phase_discrimination,*,reset_rhos = True,
                    plot_fields = False):
        self.efield_times = times_list
        self.efields = efields_list
        self.centers = centers_list
        self.set_phase_discrimination(phase_discrimination)
        self.dts = []
        self.efield_frequencies = []
        if reset_rhos:
            self.rhos = dict()
        for t in times_list:
            if t.size == 1:
                dt = 1
                w = np.array([0])
            else:
                dt = t[1] - t[0]
                w = fftshift(fftfreq(t.size,d=dt))*2*np.pi
            self.dts.append(dt)
            self.efield_frequencies.append(w)

        self.dt = self.dts[0]

        if self.detection_type == 'polarization':
            try:
                self.local_oscillator = self.efields[-1].copy()
            except:
                self.local_oscillator = copy.deepcopy(self.efields[-1])

        for field in self.efields:
            if len(field) == 1:
                # M = 1 is the impulsive limit
                pass
            else:
                self.check_efield_resolution(field,plot_fields = plot_fields)

    def check_efield_resolution(self,efield,*,plot_fields = False):
        efield_tail = np.max(np.abs([efield[0],efield[-1]]))


        if efield_tail > np.max(np.abs(efield))/100:
            warnings.warn('Consider using larger num_conv_points, pump does not decay to less than 1% of maximum value in time domain')
            
        efield_fft = fftshift(fft(ifftshift(efield)))*self.dt
        efield_fft_tail = np.max(np.abs([efield_fft[0],efield_fft[-1]]))
        
        if efield_fft_tail > np.max(np.abs(efield_fft))/100:
            warnings.warn('''Consider using smaller value of dt, pump does not decay to less than 1% of maximum value in frequency domain''')

        if plot_fields:
            fig, axes = plt.subplots(1,2)
            l1,l2, = axes[0].plot(self.efield_t,np.real(efield),self.efield_t,np.imag(efield))
            plt.legend([l1,l2],['Real','Imag'])
            axes[1].plot(self.efield_w,np.real(efield_fft),self.efield_w,np.imag(efield_fft))

            axes[0].set_ylabel('Electric field Amp')
            axes[0].set_xlabel('Time ($\omega_0^{-1})$')
            axes[1].set_xlabel('Frequency ($\omega_0$)')

            fig.suptitle('Check that efield is well-resolved in time and frequency')
            plt.show()

    def set_local_oscillator_phase(self,phase):
        self.efields[-1] = np.exp(1j*phase) * self.local_oscillator

    def get_closest_index_and_value(self,value,array):
        """Given an array and a desired value, finds the closest actual value
stored in that array, and returns that value, along with its corresponding 
array index
"""
        index = np.argmin(np.abs(array - value))
        value = array[index]
        return index, value

    def load_H_eigensystem(self):
        """Look for the eigenvalues and eigenvectors of H, and load them 
            if available. Useful as a convenience for assigning spectral 
            positions"""
        parent_dir = os.path.split(self.base_path)[0]
        H_e_name = os.path.join(parent_dir,'closed','eigenvalues.npz')
        H_ev_name = os.path.join(parent_dir,'closed','eigenvectors.npz')
        try:
            with np.load(H_e_name) as eigval_archive:
                manifolds = list(eigval_archive.keys())
                self.H_eigenvalues = {key:eigval_archive[key] for key in manifolds}
        except FileNotFoundError:
            pass
        try:
            with np.load(H_ev_name) as eigvec_archive:
                manifolds = list(eigvec_archive.keys())
                self.H_eigenvectors = {key:eigvec_archive[key] for key in manifolds}
        except FileNotFoundError:
            pass

    def load_L(self):
        """Load in known eigenvalues. Must be stored as a numpy archive file,
with keys: GSM, SEM, and optionally DEM.  The eigenvalues for each manifold
must be 1d arrays, and are assumed to be ordered by increasing energy. The
energy difference between the lowest energy ground state and the lowest 
energy singly-excited state should be set to 0
"""
        L_save_name = os.path.join(self.base_path,'L.npz')
        try:
            with np.load(L_save_name,allow_pickle=True) as L_archive:
                self.L = dict()
                for key in L_archive.keys():
                    L = L_archive[key]
                    if L.dtype == np.dtype('O'):
                        self.L[key] = L[()]
                    else:
                        if self.check_sparsity(L):
                            self.L[key] = csr_matrix(L)
                        else:
                            self.L[key] = L
        except:
            self.L = {'all_manifolds':load_npz(L_save_name)}
        self.manifolds = list(self.L.keys())

    def check_sparsity(self,mat):
        csr_mat = csr_matrix(mat)
        sparsity = csr_mat.nnz / (csr_mat.shape[0]*csr_mat.shape[1])
        if sparsity < self.sparsity_threshold:
            return True
        else:
            return False
        
    def dL(self,t,rho):
        try:
            L = self.L['all_manifolds']
        except KeyError:
            L = self.L[rho.manifold_key]
        return L.dot(rho)

    def get_dL_manual(self,manifold_key):
        try:
            L = self.L['all_manifolds']
        except KeyError:
            L = self.L[manifold_key]

        def L_fun(t,rho):
            return L.dot(rho)

        return L_fun

    def one_time_step_function(self,rho0,t0,tf,*,manifold_key = None):
        num_steps = 0
        if manifold_key == None:
            rk45 = RK45(self.dL,t0,rho0,tf,atol=self.atol,rtol=self.rtol)
        else:
            dL = self.get_dL_manual(manifold_key)
            rk45 = RK45(dL,t0,rho0,tf,atol=self.atol,rtol=self.rtol)
        while rk45.t < tf:
            rk45.step()
            num_steps += 1
        rho_final = rk45.y
        return rho_final

    def get_bottom_eigenvector(self):
        try:
            L = self.L['all_manifolds']
        except KeyError:
            L = self.L['00']
        if L.shape == (1,1):
            e = L[0,0]
            ev = np.array([[1]])
        else:
            e, ev = eigs(L,k=1,which='SM',maxiter=10000)
        if e.size == 1 and np.allclose(e,0):
            pass
        else:
            raise Exception('Smallest magnitude eigenvalue of L is {}. L must have a single stationary state for this code to work'.format(e))
        v = ev[:,0]
        H_size = int(np.sqrt(v.size))
        rho = v.reshape((H_size,H_size))
        trace = rho.trace()
        v = v/trace # Need to start with a trace 1 object
        return v

    def set_rho0_auto(self):
        try:
            rho0 = np.load(os.path.join(self.base_path,'rho0.npy'))
        except FileNotFoundError:
            rho0 = self.get_bottom_eigenvector()
        t = np.array([-np.inf,0,np.inf])
        rho0 = rho0[:,np.newaxis] * np.ones((rho0.size,t.size))
        pulse_number = None
        manifold_key = '00'
        self.rho0 = RK_rho_container(t,rho0,pulse_number,manifold_key,
                                     interp_kind = 'zero',optical_gap = self.optical_gap)

    def set_rho_shapes(self):
        self.rho_shapes = dict()
        if 'all_manifolds' in self.manifolds:
            L_size = self.L['all_manifolds'].size
            H_size = int(np.sqrt(L_size))
            self.rho_shapes['all_manifolds'] = (H_size,H_size)
        else:
            H_sizes = dict()
            for key in self.manifolds:
                ket_key, bra_key = key
                if ket_key == bra_key:
                    L_size = self.L[key].shape[0]
                    H_size = int(np.sqrt(L_size))
                    H_sizes[ket_key] = H_size
            for key in self.manifolds:
                ket_key, bra_key = key
                ket_size = H_sizes[ket_key]
                bra_size = H_sizes[bra_key]
                self.rho_shapes[key] = (ket_size,bra_size)

    def load_mu(self):
        """Load the precalculated dipole overlaps.  The dipole operator must
            be stored as a .npz file, and must contain at least one array, each with three 
            indices: (new manifold index, old manifold eigenfunction, 
            cartesian coordinate)."""
        try:
            file_name = os.path.join(self.base_path,'mu_site_basis.npz')
            with np.load(file_name) as mu_archive:
                self.mu = {key:mu_archive[key] for key in mu_archive.keys()}
        except FileNotFoundError:
            try:
                file_name = os.path.join(self.base_path,'mu_original_L_basis.npz')
                with np.load(file_name) as mu_archive:
                    self.mu = {key:mu_archive[key] for key in mu_archive.keys()}
            except FileNotFoundError:
                file_name = os.path.join(self.base_path,'mu.npz')
                with np.load(file_name) as mu_archive:
                    self.mu = {key:mu_archive[key] for key in mu_archive.keys()}
        sparse_flags = []
        for key in self.mu.keys():
            mu_2D = np.sum(np.abs(self.mu[key])**2,axis=-1)
            sparse_flags.append(self.check_sparsity(mu_2D))
        sparse_flags = np.array(sparse_flags)
        if np.allclose(sparse_flags,True):
            self.sparse_mu_flag = True
        else:
            self.sparse_mu_flag = False

        for key in self.mu.keys():
            mu_x = self.mu[key][...,0]
            mu_y = self.mu[key][...,1]
            mu_z = self.mu[key][...,2]

            if self.sparse_mu_flag:
                self.mu[key] = [csr_matrix(mu_x),csr_matrix(mu_y),csr_matrix(mu_z)]
            else:
                self.mu[key] = [mu_x,mu_y,mu_z]

        print('RKE_sparse_mu_flag',self.sparse_mu_flag)
        
    ### Setting the electric field to be used

    def set_polarization_sequence(self,polarization_list,*,reset_rhos=True):
        """Sets the sequences used for either parallel or crossed pump and probe
        
        Args:
            polarization_list (list): list of four strings, can be 'x','y' or 'z'
        Returns:
            None: sets the attribute polarization sequence
"""

        x = np.array([1,0,0])
        y = np.array([0,1,0])
        z = np.array([0,0,1])
        pol_options = {'x':x,'y':y,'z':z}

        self.polarization_sequence = [pol_options[pol] for pol in polarization_list]

        if reset_rhos:
            self.rhos = dict()


    ### Tools for recursively calculating perturbed density maatrices using TDPT

    def dipole_matrix(self,pulse_number,key,ket_flag=True,up_flag=True):
        """Calculates the dipole matrix given the electric field polarization vector,
            if ket_flag = False then uses the bra-interaction"""
        t0 = time.time()
        pol = self.polarization_sequence[pulse_number]

        x = np.array([1,0,0])
        y = np.array([0,1,0])
        z = np.array([0,0,1])
        try:
            mu = self.mu[key]
        except KeyError:
            if ket_flag:
                key = 'ket'
            else:
                key = 'bra'
            if up_flag:
                key += '_up'
            else:
                key += '_down'
            mu = self.mu[key]
            
        if np.all(pol == x):
            overlap_matrix = mu[0]#.copy()
        elif np.all(pol == y):
            overlap_matrix = mu[1]#.copy()
        elif np.all(pol == z):
            overlap_matrix = mu[2]#.copy()
        else:
            overlap_matrix = mu[0]*pol[0] + mu[1]*pol[1] + mu[2]*pol[2]

        # if self.sparse_mu_flag:
        #     to_return = csr_matrix(overlap_matrix)
        # else:
        #     to_return = overlap_matrix

        t1 = time.time()
        self.dipole_time += t1-t0

        return overlap_matrix

    def manifold_key_to_array(self,key):
        """Key must be a string of exactly 2 integers, the first describing
            the ket manifold, the second the bra manifold.  If the density 
            matrix is represented in the full space, rather than being divided
            into manifolds, the first integer reperesents the total number of
            excitations to the ket side, and the second integers represents 
            the sum of all excitations to the bra side."""
        if len(key) != 2:
            raise Exception('manifold key must be a string of exactly two intgers')
        return np.array([int(char) for char in key],dtype=int)
    
    def manifold_array_to_key(self,manifold):
        """Inverse of self.manifold_key_to_array"""
        if manifold.size != 2 or manifold.dtype != int:
            raise Exception('manifold array must contain exactly 2 integer') 
        return str(manifold[0]) + str(manifold[1])

    def next_order(self,rho_in,*,ket_flag=True,up_flag=True,pulse_number = 0):
        """This function connects psi_p to psi+pj^(*) using the Euler Method.

        Args:
            rho_in (rho_container): input density matrix
            pulse_number (int): index of optical pulse (0,1,2,...)
        
        Return:
            rho_dict (rho_container): next-order density matrix
"""     
        pulse_time = self.pulse_times[pulse_number]
        t = self.efield_times[pulse_number] + pulse_time
        old_manifold_key = rho_in.manifold_key
        if up_flag:
            change = 1
        else:
            change = -1
        if ket_flag:
            manifold_change = np.array([change,0],dtype=int)
        else:
            manifold_change = np.array([0,change],dtype=int)
        old_manifold = self.manifold_key_to_array(old_manifold_key)
        new_manifold = old_manifold + manifold_change
        new_manifold_key = self.manifold_array_to_key(new_manifold)
        mu_key = old_manifold_key + '_to_' + new_manifold_key
        
        if ket_flag == up_flag:
            # Rotating term excites the ket and de-excites the bra
            conjugate_flag = False
        else:
            # Counter-rotating term
            conjugate_flag = True

        if conjugate_flag:
            center = -self.centers[pulse_number]
        else:
            center = self.centers[pulse_number]

        M = t.size
        old_rho = rho_in(t)

        if self.conserve_memory:
            # move back to the basis the Liouvillian was written in
            if 'all_manifolds' in self.manifolds:
                ket_size,bra_size = self.rho_shapes['all_manifolds']
            else:
                ket_size,bra_size = self.rho_shapes[old_manifold_key]
                
            old_rho = old_rho.reshape(ket_size,bra_size,M)

            if ket_flag:
                old_ket_key = old_manifold_key[0]
                new_ket_key = new_manifold_key[0]
                if up_flag:
                    H_mu_key = old_ket_key + '_to_' + new_ket_key
                else:
                    H_mu_key = new_ket_key + '_to_' + old_ket_key

                mu_up_flag = up_flag
            else:
                old_bra_key = old_manifold_key[1]
                new_bra_key = new_manifold_key[1]
                if up_flag:
                    H_mu_key = old_bra_key + '_to_' + new_bra_key
                else:
                    H_mu_key = new_bra_key + '_to_' + old_bra_key
                mu_up_flag = not up_flag

            overlap_matrix = self.get_H_mu(pulse_number,H_mu_key,up_flag=mu_up_flag)
            
            ta = time.time()
            if ket_flag:
                mu_old_rho = np.einsum('ij,jkl',overlap_matrix,old_rho)
            else:
                mu_old_rho = np.einsum('ijl,jk',old_rho,overlap_matrix)
            tb = time.time()
            
            rho_vec_size = mu_old_rho.shape[0]*mu_old_rho.shape[1]
            mu_old_rho = mu_old_rho.reshape(rho_vec_size,M)

        else:
            overlap_matrix = self.dipole_matrix(pulse_number,mu_key,ket_flag=ket_flag,up_flag=up_flag)
        
        
            ta = time.time()
            mu_old_rho = overlap_matrix.dot(old_rho)
            tb = time.time()
            
        self.dipole_dot_rho_time += tb - ta
        next_rho = np.zeros(mu_old_rho.shape,dtype='complex')

        if M == 1:
            next_rho[:,0] = self.efields[pulse_number] * mu_old_rho
        else:
            if conjugate_flag:
                efield = self.efields[pulse_number]*np.exp(-1j*center*t)
            else:
                efield = np.conjugate(self.efields[pulse_number])*np.exp(-1j*center*t)


            ############
            # This 1j vs -1j needs to be derived!!! #####
            ############
            if ket_flag:
                efield = 1j * efield
            else:
                efield = -1j * efield
            ###########
            ###########
            
            dt = self.dts[pulse_number]
            next_rho[:,0] = efield[0] * mu_old_rho[:,0] * dt
            for i in range(1,t.size):
                rho0 = next_rho[:,i-1]
                t0 = t[i-1]
                ta = time.time()
                next_rho[:,i] = self.one_time_step_function(rho0,t0,t[i],manifold_key=new_manifold_key)

                tb = time.time()
                self.RK45_step_time += tb - ta
        
                next_rho[:,i] += efield[i] * mu_old_rho[:,i] * dt

        # # i/hbar Straight from perturbation theory
        # if ket_flag:
        #     rho *= 1j
        # else:
        #     rho *= -1j

        rho_out = RK_rho_container(t,next_rho,pulse_number,new_manifold_key,
                                   optical_gap = self.optical_gap)
        rho_out.one_time_step_function = self.one_time_step_function
    
        return rho_out
            
    def ket_up(self,rho_in,*,pulse_number = 0):
        """This method connects psi_p to psi_pj where the next order psi
            is one manifold above the current manifold.

        Args:
            rho_in (rho_container): input density matrix
            pulse_number (int): index of optical pulse (0,1,2,...)

        Returns:
            (rho_container): output from method next_order
"""
        return self.next_order(rho_in,ket_flag=True,up_flag=True,
                               pulse_number = pulse_number)

    def ket_down(self,rho_in,*,pulse_number = 0):
        """This method connects psi_p to psi_pj where the next order psi
            is one manifold above the current manifold.

        Args:
            rho_in (rho_container): input density matrix
            pulse_number (int): index of optical pulse (0,1,2,...)

        Returns:
            (rho_container): output from method next_order
"""
        return self.next_order(rho_in,ket_flag=True,up_flag=False,
                               pulse_number = pulse_number)

    def bra_up(self,rho_in,*,pulse_number = 0):
        """This method connects psi_p to psi_pj where the next order psi
            is one manifold above the current manifold.

        Args:
            rho_in (rho_container): input density matrix
            pulse_number (int): index of optical pulse (0,1,2,...)

        Returns:
            (rho_container): output from method next_order
"""
        return self.next_order(rho_in,ket_flag=False,up_flag=True,
                               pulse_number = pulse_number)

    def bra_down(self,rho_in,*,pulse_number = 0):
        """This method connects psi_p to psi_pj where the next order psi
            is one manifold above the current manifold.

        Args:
            rho_in (rho_container): input density matrix
            pulse_number (int): index of optical pulse (0,1,2,...)

        Returns:
            (rho_container): output from method next_order
"""
        return self.next_order(rho_in,ket_flag=False,up_flag=False,
                               pulse_number = pulse_number)

    ### Tools for taking the expectation value of the dipole operator with perturbed density matrices

    def dipole_down_H_mu(self,rho,manifold_key,*,new_manifold_mask = None,pulse_number = -1,
                    ket_flag=True):
        """This method is similar to the method down, but does not involve 
            the electric field shape or convolutions. It is the action of the 
            dipole operator on the ket-side without TDPT effects.  It also includes
            the dot product of the final electric field polarization vector."""
        
        if not ket_flag:
            raise Exception('Not implemented for bra-side')
        old_manifold_key = manifold_key
        old_ket_key = old_manifold_key[0]
        new_ket_key = str(int(old_ket_key)-1)
        mu_key = new_ket_key + '_to_' + old_ket_key

        if ket_flag:
            center = - self.centers[pulse_number]
            conjugate_flag = True
        else:
            center = self.centers[pulse_number]
            conjugate_flag = False

        t_size = rho.shape[-1]
        
        if 'all_manifolds' in self.L.keys():
            L_size = rho.size
            H_size = int(np.sqrt(L_size))
            rho = rho.reshape(H_size,H_size,t_size)
        else:
            ket_manifold_key = old_manifold_key[0] + old_manifold_key[0]
            ket_L_manifold_size = self.L[ket_manifold_key].shape[0]
            ket_H_size = int(np.sqrt(ket_L_manifold_size))

            bra_manifold_key = old_manifold_key[1] + old_manifold_key[1]
            bra_L_manifold_size = self.L[bra_manifold_key].shape[0]
            bra_H_size = int(np.sqrt(bra_L_manifold_size))

            rho = rho.reshape(ket_H_size,bra_H_size,t_size)
        
        overlap_matrix = self.get_H_mu(pulse_number,mu_key,ket_flag=ket_flag,up_flag=False)

        t0 = time.time()
        polarization_field = np.einsum('ij,jik',overlap_matrix,rho)
                
        t1 = time.time()

        return polarization_field

    def dipole_down_L_mu(self,rho,manifold_key,*,new_manifold_mask = None,
                         pulse_number = -1,ket_flag=True):
        """This method is similar to the method down, but does not involve 
            the electric field shape or convolutions. It is the action of the 
            dipole operator on the ket-side without TDPT effects.  It also includes
            the dot product of the final electric field polarization vector."""
        old_manifold_key = manifold_key
        change = -1
        if ket_flag:
            manifold_change = np.array([change,0])
        else:
            manifold_change = np.array([0,change])

        old_manifold = self.manifold_key_to_array(old_manifold_key)
        new_manifold = old_manifold + manifold_change
        new_manifold_key = self.manifold_array_to_key(new_manifold)
        mu_key = old_manifold_key + '_to_' + new_manifold_key
        
        if ket_flag:
            center = - self.centers[pulse_number]
            conjugate_flag = True
        else:
            center = self.centers[pulse_number]
            conjugate_flag = False

        rho_in = rho
        
        overlap_matrix = self.dipole_matrix(pulse_number,mu_key,ket_flag=True,up_flag=False)

        t0 = time.time()
        rho = overlap_matrix.dot(rho_in)
                
        t1 = time.time()

        L_size = rho.shape[0]
        H_size = int(np.sqrt(L_size))

        # reshape rho into a normal density matrix representation
        rho = rho.reshape((H_size,H_size,rho.shape[-1]))

        polarization_field = np.einsum('iij',rho)

        return polarization_field

    def load_H_mu(self):
        parent_dir = os.path.split(self.base_path)[0]
        file_name = os.path.join(parent_dir,'closed','mu.npz')

        with np.load(file_name) as mu_archive:
            self.H_mu = {key:mu_archive[key] for key in mu_archive.keys()}

    def get_H_mu(self,pulse_number,key,ket_flag=True,up_flag=True):
        """Calculates the dipole matrix given the electric field polarization vector,
            if ket_flag = False then uses the bra-interaction"""
        t0 = time.time()
        pol = self.polarization_sequence[pulse_number]

        x = np.array([1,0,0])
        y = np.array([0,1,0])
        z = np.array([0,0,1])
        try:
            mu = self.H_mu[key]
        except KeyError:
            try:
                key = 'up'
                mu = self.H_mu[key]
            except KeyError:
                key = 'ket_up'
                mu = self.H_mu[key]
            
        if np.all(pol == x):
            overlap_matrix = mu[:,:,0].copy()
        elif np.all(pol == y):
            overlap_matrix = mu[:,:,1].copy()
        elif np.all(pol == z):
            overlap_matrix = mu[:,:,2].copy()
        else:
            overlap_matrix = np.tensordot(mu,pol,axes=(-1,0))

        if not up_flag:
            overlap_matrix = overlap_matrix.T

        t1 = time.time()
        self.dipole_time += t1-t0

        return overlap_matrix

    def set_undersample_factor(self,frequency_resolution):
        """dt is set by the pulse. However, the system dynamics may not require such a 
            small dt.  Therefore, this allows the user to set a requested frequency
            resolution for any spectrally resolved signals."""
        # f = pi/dt
        dt = np.pi/frequency_resolution
        u = int(np.floor(dt/self.dt))
        self.undersample_factor = max(u,1)
        
    def dipole_expectation(self,rho_in,*,pulse_number = -1,ket_flag=True):
        """Computes the expectation value of the dipole operator"""
        t0 = time.time()

        pulse_number = -1
        
        pulse_time = self.pulse_times[pulse_number]
        
        efield_t = self.efield_times[pulse_number] + pulse_time

        if ket_flag:
            center = - self.centers[pulse_number]
        else:
            center = self.centers[pulse_number]
        
        # The signal is zero before the final pulse arrives, and persists
        # until it decays. Therefore we avoid taking the sum at times
        # where the signal is zero.
        t = self.t + pulse_time

        pulse_start_ind = np.argmin(np.abs(t-efield_t[0]))
        if efield_t[0] < t[pulse_start_ind]:
            pulse_start_ind -= 1
        pulse_end_ind = np.argmin(np.abs(t-efield_t[-1]))
        if efield_t[-1] > t[pulse_end_ind]:
            pulse_end_ind += 1
        
        t_slice = slice(pulse_start_ind,None,None)
        t1_slice = slice(pulse_start_ind,pulse_end_ind,None)
        u = self.undersample_factor
        t2_slice = slice(pulse_end_ind,None,u)

        t = self.t[t_slice] + pulse_time
        t1 = self.t[t1_slice] + pulse_time
        t2 = self.t[t2_slice] + pulse_time

        rho1 = rho_in(t1)
        rho2 = rho_in(t2)

        rho1 *= np.exp(-1j * center * t1[np.newaxis,:])

        rho2 *= np.exp(-1j * center * t2[np.newaxis,:])

        # _u is an abbreviation for undersampled
        t_u = np.hstack((t1,t2))
        rho_u = np.hstack((rho1,rho2))
        
        tb = time.time()
        self.slicing_time += tb-t0

        t0 = time.time()
        exp_val_u = self.dipole_down(rho_u,rho_in.manifold_key,pulse_number = pulse_number,
                                     ket_flag = ket_flag)
        
        tb = time.time()
        self.expectation_time += tb-t0

        t0 = time.time()

        # Interpolate expectation value back onto the full t-grid
        if u != 1:
            # Often must extrapolate the final point
            exp_val_interp = scipy.interpolate.interp1d(t_u,exp_val_u,kind='cubic',fill_value='extrapolate')
            exp_val = exp_val_interp(t)
        else:
            exp_val = exp_val_u
        # print(exp_val.size/exp_val_u.size)
        tb = time.time()
        self.interpolation_time += tb-t0
        
        # Initialize return array with zeros
        ret_val = np.zeros(self.t.size,dtype='complex')
        
        # set non-zero values using t_slice
        ret_val[pulse_start_ind:] = exp_val
        return ret_val

    def integrated_dipole_expectation(self,rho_in,*,ket_flag=True):
        """Computes the expectation value of the dipole operator"""
        
        pulse_number = -1
        
        pulse_time = self.pulse_times[pulse_number]
        t = pulse_time + self.efield_times[pulse_number]

        if ket_flag:
            center = - self.centers[pulse_number]
        else:
            center = self.centers[pulse_number]

        rho = rho_in(t)

        rho_nonzero = rho_in.bool_mask
        try:
            ev = self.eigenvalues['all_manifolds'][rho_nonzero]
        except KeyError:
            ev = self.eigenvalues[rho_in.manifold_key][rho_nonzero]

        rho = rho * np.exp((ev[:,np.newaxis] - 1j*center)*t)

        rho_dict = {'bool_mask':rho_nonzero,'rho':rho,'manifold_key':rho_in.manifold_key}

        t0 = time.time()
        exp_val = self.dipole_down(rho_dict,pulse_number = pulse_number,
                                     ket_flag = ket_flag)
        
        tb = time.time()
        self.expectation_time += tb-t0

        return exp_val

    def get_local_oscillator(self):
        local_oscillator_number = -1
        efield_t = self.efield_times[local_oscillator_number]
        efield = self.efields[local_oscillator_number]

        if efield_t.size == 1:
            # Impulsive limit: delta in time is flat in frequency
            efield_ft = np.ones(self.w.size)*efield
            return efield_ft
        
        e_dt = efield_t[1] - efield_t[0]
        dt = self.t[1] - self.t[0]
            
        if (np.isclose(e_dt,dt) and efield_t[-1] <= self.t[-1]):
            full_efield = np.zeros(self.t.size,dtype='complex')

            # the local oscillator sets the "zero" on the clock
            pulse_time_ind = np.argmin(np.abs(self.t))

            pulse_start_ind = pulse_time_ind - efield_t.size//2
            pulse_end_ind = pulse_time_ind + efield_t.size//2 + efield_t.size%2

            t_slice = slice(pulse_start_ind, pulse_end_ind,None)
            
            full_efield[t_slice] = efield
            efield_ft = fftshift(ifft(ifftshift(full_efield)))*full_efield.size * dt
        else:
            efield_ft = fftshift(ifft(ifftshift(efield))) * efield.size * e_dt
            efield_w = fftshift(fftfreq(efield_t.size,d=e_dt)) * 2 * np.pi
            fill_value = (efield_ft[0],efield_ft[-1])
            f = sinterp1d(efield_w,efield_ft,fill_value = fill_value,
                          bounds_error=False,kind='quadratic')
            efield_ft = f(self.w)

        return efield_ft
    
    def polarization_to_signal(self,P_of_t_in,*,
                                local_oscillator_number = -1,undersample_factor = 1):
        """This function generates a frequency-resolved signal from a polarization field
           local_oscillator_number - usually the local oscillator will be the last pulse 
                                     in the list self.efields"""
        undersample_slice = slice(None,None,undersample_factor)
        P_of_t = P_of_t_in[undersample_slice].copy()
        t = self.t[undersample_slice]
        dt = t[1] - t[0]
        pulse_time = self.pulse_times[local_oscillator_number]
        efield_t = self.efield_times[local_oscillator_number]
        
        pulse_time_ind = np.argmin(np.abs(self.t))
        
        efield = self.get_local_oscillator()

        halfway = self.w.size//2
        pm = self.w.size//(2*undersample_factor)
        efield_min_ind = halfway - pm
        efield_max_ind = halfway + pm + self.w.size%2
        efield = efield[efield_min_ind:efield_max_ind]

        P_of_w = fftshift(ifft(ifftshift(P_of_t)))*len(P_of_t)*dt#/np.sqrt(2*np.pi)

        signal = P_of_w * np.conjugate(efield)
        if not self.return_complex_signal:
            return np.imag(signal)
        else:
            return -1j*signal

    def integrated_polarization_to_signal(self,P,*,
                                local_oscillator_number = -1):
        """This function generates a frequency-resolved signal from a polarization field
           local_oscillator_number - usually the local oscillator will be the last pulse 
                                     in the list self.efields"""
        efield_t = self.efield_times[local_oscillator_number]

        efield = self.efields[local_oscillator_number]

        signal = np.trapz(P * np.conjugate(efield),x=efield_t)
        if not self.return_complex_signal:
            return np.imag(signal)
        else:
            return -1j*signal

    def save(self,file_name,pulse_delay_names,*,use_base_path=True):
        if use_base_path:
            file_name = os.path.join(self.base_path,file_name)
        save_dict = {}
        for name,delays in zip(pulse_delay_names,self.all_pulse_delays):
            save_dict[name] = delays
        if self.detection_type == 'polarization':
            save_dict['wt'] = self.w
        save_dict['signal'] = self.signal
        save_dict['signal_calculation_time'] = self.calculation_time
        np.savez(file_name,**save_dict)



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider
import matplotlib
from scipy.sparse import csr_matrix, kron
from scipy.sparse.linalg import eigs, eigsh
import itertools
from scipy.linalg import eig, expm, eigh
import yaml
import os
from scipy.special import factorial
from numpy.polynomial.hermite import hermval
import warnings
import time

class LadderOperators:
    """Constructs creation and annihilation operators for the SHO.
    
    This class is used to construct harmonic and anharmonic Hamiltonians.  
    Powers of x and p are created using ladder operators that are larger than
    the desired size of the truncated matrices, in order to avoid errors

    Attributes:
        size (int): size of returned operators
        calculation_size (int): size of internal operators
        a (np.ndarray): annihilation operator
        ad (np.ndarray): creation operator
        x (np.ndarray): position operator
        p (np.ndarray); momentum operator
        
"""
    def __init__(self,size,*,disp = 0, extra_size=10):
        """
        Args:
            size (int): desired size of returned operators
            disp (complex): complex phase-space displacement of the bottom
                of the potential energy surface
            extra_size (int): additional size used for calculating 
                matrix powers default value 10
"""
        self.size = size
        self.disp = disp
        self.calculation_size = size + extra_size
        self.set_a_and_ad()
        self.set_x_and_p()
        self.set_N()

    def destroy(self):
        """Constructs annihilation operator for dispalced SHO

        Returns:
            np.ndarray: annihilation operator
"""
        def offdiag1(n):
            return np.sqrt(n+1)

        n = np.arange(0,self.calculation_size)
        off1 = offdiag1(n[0:-1])
        a = - np.eye(self.calculation_size) * self.disp/np.sqrt(2)
        a += np.diag(off1,k=1)
        return a

    def create(self):
        """Constructs creation operator for displaced SHO

        Returns:
            np.ndarray: creation operator
"""
        a = self.destroy()
        ad = np.conjugate(a.transpose())
        return ad

    def set_a_and_ad(self):
        """Sets raising (a^dagger) and lowering (a) operators
"""
        self._a = self.destroy()
        self.a = self._a[:self.size,:self.size]

        self._ad = self.create()
        self.ad = self._ad[:self.size,:self.size]

    def set_N(self):
        """Constructs number operator"""
        self._N = self._ad.dot(self._a)
        self.N = self._N[:self.size,:self.size]

    def set_x_and_p(self):
        """Sets the position (x) and momentum (p) operators using the 
            creation and annihilation operator methods.
"""
        self._x = (self._ad + self._a)/np.sqrt(2)
        self.x = self._x[:self.size,:self.size]
        
        self._p = 1j* (self._ad - self._a)/np.sqrt(2)
        self.p = self._p[:self.size,:self.size]

    def x_power_n(self,n):
        """Calculates x^n 

        Args:
            n (int): power

        Returns:
            np.ndarray: position operator x to the n power
"""
        xn = np.linalg.matrix_power(self._x,n)
        return xn[:self.size,:self.size]

    def p_power_n(self,n):
        """Calculates p^n 
        
        Args:
            n (int): power

        Returns:
            np.ndarray: momentum operator p to the n power
"""
        pn = np.linalg.matrix_power(self._p,n)
        return pn[:self.size,:self.size]
    
    def _displacement_operator(self,alpha):
        """Creates displacement operator of size N x N where N is the 
            calculation size
        
        Args:
            alpha (complex) : complex displacement
        Returns:
            np.ndarray : displacement operator
        """
        return expm(alpha * self._ad - np.conjugate(alpha) * self._a)
    
    def displacement_operator(self,alpha):
        """Creates displacement operator of size N x N where N is the size
        
        Args:
            alpha (complex) : complex displacement
        Returns:
            np.ndarray : displacement operator
        """
        return self._displacement_operator(alpha)[:self.size,:self.size]
    
    def _squeezing_operator(self,zeta):
        """Creates squeezing operator of size N x N where N is the 
            calculation size
        
        Args:
            zeta (complex) : complex squeezing parameter
        Returns:
            np.ndarray : squeezing operator
        """
        a2 = np.linalg.matrix_power(self._a,2)
        ad2 = np.linalg.matrix_power(self._ad,2)
        return expm((np.conjugate(zeta) * a2 - zeta * ad2)/2)
    
    def squeezing_operator(self,zeta):
        """Creates squeezing operator of size N x N where N is the size
        
        Args:
            zeta (complex) : complex squeezing parameter
        Returns:
            np.ndarray : squeezing operator
        """
        return self._squeezing_operator(zeta)[:self.size,:self.size]
    
class DisplacedAnharmonicOscillator:
    """Hamiltonian for a displaced anharmonic oscillator
        
    This class can model all the same things that ArbitraryHamiltonian can.  
    However, it is much easier to work with when you need to displace an 
    anharmonic oscillator in x

    Attributes:
        constant (float): constant energy offset
        displacement (complex): complex phase-space displacement of the 
            bottom of the potential energy surface
        pcoef (list): coefficients for 2nd order and higher momentum terms
        xcoef (list): coefficients for 2nd order and higher position terms
        ham (np.ndarray): Hamiltonian
"""
    def __init__(self,size):
        """
        Args:
            size (int) : truncation size for Hamiltonian
"""
        self.size = size

    def set_ham(self,constant,displacement,pcoef,xcoef):
        """Creates anharmonic Hamiltonian for 1 vibrational mode

        Args:
            constant (complex) : complex reorganization energy
            displacement (complex) : phase-space displacement
            pcoef (list): coefficients for 2nd order + higher momentum terms
            xcoef (list): coefficients for 2nd order + higher position terms
"""
        extra_size = max(len(pcoef),len(xcoef))
        L = LadderOperators(self.size,disp=displacement,extra_size=extra_size)
        self.constant = constant
        self.displacement = displacement
        self.pcoef = pcoef
        self.xcoef = xcoef
        L.set_x_and_p()
        self.ham = np.eye(self.size,dtype=complex)*constant
        for n in range(0,len(pcoef)):
            self.ham += pcoef[n] * L.p_power_n(n+2)
        for m in range(0,len(xcoef)):
            self.ham += xcoef[m] * L.x_power_n(m+2)
            
        if np.allclose(np.imag(self.ham),0):
            self.ham = np.real(self.ham)

    def potential(self,x):
        """Potential energy surface for this Hamiltonian
        
        Args:
            x (np.ndarray): 1d-array of position values

        Returns:
            np.ndarray: 1d-array of potential energy as a function of x
"""
        y = np.ones(x.shape,x.dtype)*self.constant
        x0 = np.real(self.displacement)
        for i in range(len(self.xcoef)):
            y += np.power(x-x0,i+2) * self.xcoef[i]
        return y

    def plot(self,x):
        """Displays plot of eigenfunctions as a function of x

        Args:
            x (np.ndarray): 1d-array of position values
"""
        fig, ax = plt.subplots()
        ax.plot(x,self.potential(x))
        for n in range(len(self.eigvals)):
            ev = self.eigvals[n]
            vec = self.eigvecs[:,n]
            ax.axhline(y=ev,linestyle='--')
            f = oscillator_1d(vec,x)
            ax.plot(x,f**2+ev)

    def set_eig(self,*,hermitian=True):
        """Use eigensolver to find all eigenvalues

        Keyword Args:
            hermitian (bool) : True if self.ham is Hermitian
"""
        if hermitian:
            eigvals, eigvecs = eigh(self.ham)
        else:
            eigvals, eigvecs = eig(self.ham)
        sort_indices = eigvals.argsort()
        eigvals.sort()
        eigvecs = eigvecs[:,sort_indices]
        # I choose to pick the phase of my eigenvectors such that the state
        # which has the largest overlap has a positive overlap. For
        # sufficiently small d, and alpha close to 1, this will be the
        # overlap between the same excited and ground states.
        for i in range(eigvals.size):
            max_index = np.argmax(np.abs(eigvecs[:,i]))
            if eigvecs[max_index,i] < 0:
                eigvecs[:,i] *= -1
        self.eigvecs = eigvecs
        if np.max(np.imag(eigvals)) > 1E-12:
            warnings.warn('Eigenvalues are not fully real')
            self.eigvals = eigvals
        else:
            self.eigvals = np.real(eigvals)

    def set_eigs(self,num,*,hermitian=True):
        """Use sparse eigensolver to find the bottom specified eigenvalues

        Args:
            num (int): number of eigenvalues to find
        Keyword Args:
            hermitian (bool) : True if self.ham is Hermitian
"""
        if hermitian:
            eigvals, eigvecs = eigsh(self.ham,k=num,which='SM')
        else:
            eigvals, eigvecs = eigs(self.ham,k=num,which='SM')
        sort_indices = eigvals.argsort()
        eigvals.sort()
        eigvecs = eigvecs[:,sort_indices]
        # I choose to pick the phase of my eigenvectors such that the state
        # which has the largest overlap has a positive overlap. For
        # sufficiently small d, and alpha close to 1, this will be the
        # overlap between the same excited and ground states.
        for i in range(num):
            max_index = np.argmax(np.abs(eigvecs[:,i]))
            if eigvecs[max_index,i] < 0:
                eigvecs[:,i] *= -1
        self.eigvecs = eigvecs
        if np.max(np.imag(eigvals)) > 1E-12:
            warnings.warn('Eigenvalues are not fully real')
            self.eigvals = eigvals
        else:
            self.eigvals = np.real(eigvals)

    def displaced_squeezed_state(self,state,alpha,zeta):
        """Takes in a wavefunction and applies first the squeeze operator,
            followed by the displacement operator
        
        Args:
            state (np.ndarray) : wavefunction
            alpha (complex) : complex displacement
            zeta (complex) : complex squeezing
            
        Returns:
            (np.ndarray) : displaced and squeezed wavefunction
        """
        L = LadderOperators(self.size,disp=0)
        D = L.displacement_operator(alpha)
        S = L.squeezing_operator(zeta)
        return D.dot(S.dot(state))

    def displaced_vacuum(self,*,disp = 1):
        """Create a displaced vacuum state relative to an unshifted potential
"""
        vac = np.zeros(self.size)
        vac[0] = 1
        return self.displacced_squeezed_state(vac,disp,0)

    def time_evolution(self,state,time):
        """Evolves a state forward in time using the eigensystem
"""
        basis_change = self.eigvecs
        state = np.conjugate(basis_change.T).dot(state)
        time_oper = np.exp(-1j * self.eigvals * time)
        state = state * time_oper
        state = basis_change.dot(state)
        return state

    def animate_wavepacket(self,state,*,x=np.arange(-3,3,0.01),
                           t=np.arange(0,2*np.pi,np.pi/30)):
        fig, ax = plt.subplots()
        ax.plot(x,self.potential(x))
        fig.subplots_adjust(bottom=0.4)
        t_ax = fig.add_axes([0.1,0.05,0.65,0.02])
        t_slider = Slider(t_ax, 'time', 0, t[-1],valinit=0)

        psi = oscillator_1d(state,x)

        line_r, line_i, line_abs, = ax.plot(x,np.real(psi)*2+2,
                                            x,np.imag(psi)*2+2,
                                            x,np.abs(psi)*2+2)
        
        def update(val):
            tval = t_slider.val
            psi = oscillator_1d(self.time_evolution(state,tval),x)
            line_r.set_ydata(np.real(psi)*2+2)
            line_i.set_ydata(np.imag(psi)*2+2)
            line_abs.set_ydata(np.abs(psi)*2+2)
        t_slider.on_changed(update)

    def get_mean_x(self,t,state):
        """Calculates expectation value and uncertainty in operator x

        Args:
            t (np.ndarray) : times to calculate expectation values
            state (np.ndarray) : wavefunction at time t = 0

        Returns:
            (tuple) : expectation value of x and associated uncertainty
        """
        LO = LadderOperators(self.size)
        x = LO.x
        x2 = LO.x_power_n(2)
        return self.get_mean_and_Delta(t,state,x,op2=x2)

    def get_mean_p(self,t,state):
        """Calculates expectation value and uncertainty in operator p

        Args:
            t (np.ndarray) : times to calculate expectation values
            state (np.ndarray) : wavefunction at time t = 0

        Returns:
            (tuple) : expectation value of p and associated uncertainty
        """
        LO = LadderOperators(self.size)
        p = LO.p
        p2 = LO.p_power_n(2)
        return self.get_mean_and_Delta(t,state,p,op2=p2)
    
    def get_mean_N(self,t,state):
        """Calculates expectation value and uncertainty in operator N
            (mean number of excitations)

        Args:
            t (np.ndarray) : times to calculate expectation values
            state (np.ndarray) : wavefunction at time t = 0

        Returns:
            (tuple) : expectation value of N and associated uncertainty
        """
        LO = LadderOperators(self.size)
        N = LO.N
        N2 = np.dot(LO._N,LO._N)[:self.size,:self.size]
        return self.get_mean_and_Delta(t,state,N,op2=N2)
    
    def get_mean_and_Delta(self,t,state,op,*,op2='auto'):
        """Calculates expectation value and uncertainty in operator "op"

        Args:
            t (np.ndarray) : times to calculate expectation values
            state (np.ndarray) : wavefunction at time t = 0
            op (np.ndarray) : operator to calculate expectation value

        Keyword Args:
            op2 (np.ndarray) : square of the operator; if set to 'auto', 
                calculates op^2 using op
        Returns:
            (tuple) : expectation value and associated uncertainty
        """
        if type(op2) is str:
            if op2 == 'auto':
                op2 = np.dot(op,op)
        mean_op = np.zeros(t.size)
        var_op = np.zeros(t.size)
        for i in range(t.size):
            psi = self.time_evolution(state,t[i])
            op_psi = op.dot(psi)
            op2_psi = op2.dot(psi)
            mean_op[i] = np.real(np.dot(np.conjugate(psi),op_psi))
            var_op[i] = np.real(np.dot(np.conjugate(psi),op2_psi))

        Delta_op = np.sqrt(var_op - mean_op**2)

        return mean_op, Delta_op

    def plot_mean(self,t,mean,Delta):
        plt.figure()
        plt.plot(t,mean,'k')
        plt.plot(t,mean+Delta,'--k')
        plt.plot(t,mean-Delta,'--k')


class Polymer:
    """Creates the Hamiltonian and other useful operators describing a
        Polymer of two-level systems (2LS) or three-level systems (3LS)

    Attributes:
        If system is a polymer of two-level systems:
            num_sites (int): number of two level systems
            energies (list): excitation energy of each 2LS if it were isolated
            couplings (list): energetic couplings between singly-
                excited electronic states in the site basis, ordered as
                [J12,J13,...,J1N,J23,...,J2N,...]


        pols (list): set to value ['x','y','z']
        N (int): size of each 2LS (set to value 2 for 2LS)
        up (np.ndarray): raising operator for a 2LS
        down (np.ndarray): lowering operator for a 2LS
        ii (np.ndarray): identity operator for a 2LS
        occupied (np.ndarray): occupation operator for a 2LS
        empty (np.ndarray): ii - occupied

        If system is a polymer of three-level systems:
            num_sites (int) : number of three level systems
            energies (list) : one list, where each element is itself a list of two elements, the first being the
                first excitation energy 'e' of each 3LS, the second being the
                second excitation energy 'f', written as [[e1,f1],[e2,f2]...[en,fn]]
            couplings (list) : couplings between singly-excited states in the site
                basis (the J_ij coupling), between doubly excited states (K_ij) and between triply excited
                states (L_ij). They are ordered as a list of matrices instead of a single list as for 2LS
                [[J11, J12, J13,..., J1n], [J21, J22,..., J2n],..., [Jn1,..., Jn(n-1), Jnn]]
                [[K11, K12, K13,..., K1n], [K21, K22,..., K2n],..., [Kn1,..., Kn(n-1), Knn]]
                [[L11, L12, L13,..., L1n], [L21, L22,..., L2n],..., [Ln1,..., Ln(n-1), Lnn]].
                # The following rule is always true for all O= J,K,L elements: O_mn = O_nm* where the *
                is the complex conjugate

"""

    def __init__(self, site_energies, site_couplings, dipoles,*,
                 maximum_manifold='auto'):
        """This initializes an object with an arbitrary number of
        coupled two- or three-level systems, depending on N

        Args:
            site_energies (list): excitation energies of individual sites
            site_couplings (list): energetic couplings between singly-
                excited electronic states in the site basis, ordered as
                [J12,J13,...,J1N,J23,...,J2N,...]
"""

        self.num_sites = len(site_energies)

        if isinstance(site_energies[0], list):
            self.N = 3
        else:
            self.N = 2
            self.energies = site_energies
            self.couplings = site_couplings

        self.maximum_manifold = self.num_sites * (self.N - 1)

        if maximum_manifold == 'auto':
            self.set_electronic_total_occupation_number()
        else:
            self.maximum_manifold = min(maximum_manifold,self.maximum_manifold)

        if self.N == 3:
            self.energies = []
            self.double_energies = []
            for i in range(self.num_sites):
                egy = site_energies[i][0]  # list of first excitation energies
                self.energies.append(egy)
                egy = site_energies[i][1]  # list of second excitation energies
                self.double_energies.append(egy)
            # Couplings matrices for 3LS
            if self.num_sites == 1:
                self.J_list = []
                self.K_list = []
                self.L_list = []
            elif isinstance(site_couplings[0][0], list) or isinstance(site_couplings[0][0], np.ndarray):
                self.J_list = []
                self.K_list = []
                self.L_list = []
                for i in range(self.num_sites):
                    for j in range(i+1,self.num_sites):
                        self.J_list.append(site_couplings[0][i][j])
                        self.K_list.append(site_couplings[1][i][j])
                        self.L_list.append(site_couplings[2][i][j])
            else:
                self.J_list = site_couplings[0]
                self.K_list = site_couplings[1]
                self.L_list = site_couplings[2]

        self.dipoles = dipoles
        self.pols = ['x', 'y', 'z']

        ### Operators that define a single two-level system (2LS)

        self.up = self.basis(1, 0)  # a_dagger

        self.down = self.basis(0, 1)  # a

        self.ii = np.eye(self.N)

        self.occupied = self.basis(1, 1)  # this state being in excited state

        self.empty = self.basis(0, 0)

        # For a 3LS

        if self.N == 3:
            self.up10 = self.basis(1, 0)  # a_dagger

            self.up20 = self.basis(2, 0)  # b_dagger

            self.up21 = self.basis(2, 1)  # c_dagger

            self.down01 = self.basis(0,1) # a

            self.down02 = self.basis(0, 2)  # b

            self.down12 = self.basis(1, 2)  # c

            self.occupied_1 = self.basis(1, 1)

            self.occupied_2 = self.basis(2, 2)

            self.empty = self.basis(0, 0)

        ### Kron up the single 2LS operators to act on the full Hilbert space of the polymer

        self.set_up_list()
        self.set_down_list()
        self.set_occupied_list()
        self.set_empty_list()
        self.set_exchange_list()

        # For 3LS
        if self.N == 3:
            self.set_doubly_occupied_list()
            self.set_exchange_list_G()
            self.set_exchange_list_GG()
            self.set_exchange_list_C()
            self.set_up10_list()
            self.set_up21_list()
            self.set_up20_list()
            self.set_down01_list()
            self.set_down12_list()
            self.set_down02_list()

        ## Make Hamiltonian (total and by manifold)

        self.set_electronic_hamiltonian()

        self.make_mu_dict_site_basis()
        self.make_mu_site_basis('x')
        self.make_mu_up_dict_site_basis()
        self.make_mu_up_site_basis('x')
        self.make_mu_up_site_basis('y')
        self.make_mu_up_site_basis('z')

        self.make_mu_up_3d()

        self.set_manifold_eigensystems()
        self.set_electronic_eigensystem()

    def basis(self, row, col):
        """Matrix basis for a single two-level system
"""
        b = np.zeros((self.N, self.N))
        b[row, col] = 1
        return b

    ### Tools for making the basic operators in the polymer space

    def make_HI_i(self, H_i):
        self.HI = []
        for i in range(len(self.H_i)):
            self.HI_i = self.electronic_identity_kron([(self.H_i[i], i)])
            self.HI.append(self.HI_i)
        return self.HI

    def electronic_identity_kron(self, element_list):
        """Takes in a list of tuples (element, position) and krons
            them up into the full space (element is assumed to be
            a 2x2 array acting on a single 2LS.  Any positions that
            are unspecified are assumed to be identities
        Args:
            element_list (list): list of tuples

        Returns:
            2^n x 2^n array acting on the full space of the polymer
"""
        # all unspecified positions will be taken to be identities
        num_identities = self.num_sites - len(element_list)
        if num_identities < 0:
            # cannot have more operators than 2LS's
            raise ValueError('Too many elements for Hilbert space')
        # create a list of all identities
        matrix_list = [self.ii for j in range(self.num_sites)]

        for el, pos in element_list:
            # for each position specified, replace identity in
            # matrix_list with element
            matrix_list[pos] = el
        return self.recursive_kron(matrix_list)

    def recursive_kron(self, list_of_matrices):
        """Krons together a list of matrices, and removes indices corresponding to
            excitations that are above the maximum_manifold Attribute
        Args:
            list_of_matrices (list): list of np.ndarrays
        Returns:
            the kronecker product of all the input matrices, of size H_dim x H_dim,
                (H_dim is an Attribute, set by this method)
"""
        list_of_excitations = [np.arange(self.N)]*self.num_sites
        ex = list_of_excitations.pop(0)
        mat = list_of_matrices.pop(0)
        ex, mat = self.remove_higher_excitations(ex, mat)
        n = len(list_of_matrices)
        for next_item,next_ex in zip(list_of_matrices,list_of_excitations):
            mat = np.kron(mat, next_item)
            ex = self.kron_add1d(ex,next_ex)
            ex, mat = self.remove_higher_excitations(ex, mat)
        self.electronic_total_occupation_number = ex
        self.H_dim = ex.size
        return mat

    def kron_add1d(self,a,b):
        """Adds two vectors from different spaces by creating a larger space in which 
            to add them, using the kronecker product with vectors of ones
        Args:
            a (1d np.ndarray): 1d array
            b (1d np.ndarray): 1d array
        Returns:
            sum of a and b in a new space
"""
        a_ones = np.ones(a.size)
        b_ones = np.ones(b.size)
        return np.kron(a,b_ones) + np.kron(a_ones,b)

    def remove_higher_excitations(self,excitations,matrix):
        """Finds entries corresponding to excitation numbers above the maximum_manifold
            Attribute, and removes them both from the list of excitations, as well as
            from the input matrix.
        Args:
            excitations (1d np.ndarray): list of excitation number associated with each
                index of the matrix
            matrix (2d np.ndarray): a 2d square array
        Returns:
            (excitations, matrix) trimmed down to only contained the allowed number of
                excitations
"""
        inds = np.where(excitations <= self.maximum_manifold)[0]
        excitations = excitations[inds]
        matrix = matrix[inds,:]
        matrix = matrix[:,inds]
        return excitations, matrix

    def make_single_operator_list(self, O):
        """Make a list of full-space operators for a given 2x2 operator,
            by taking tensor product with identities on the other
            excitations
"""
        O_list = []
        for i in range(self.num_sites):
            Oi = self.electronic_identity_kron([(O, i)])
            O_list.append(Oi)
        return O_list

    def make_multi_operator_list(self, o_list):
        """Make a list of full-space operators for a given set of 2x2
            operators by inserting the necessary identities
"""
        O_list = []
        positions = itertools.combinations(range(self.num_sites), len(o_list))
        for pos_tuple in positions:
            Oi = self.electronic_identity_kron(list(zip(o_list, pos_tuple)))
            O_list.append(Oi)
        return O_list

    def set_occupied_list(self):
        self.occupied_list = self.make_single_operator_list(self.occupied)

    def set_doubly_occupied_list(self):
        self.doubly_occupied_list = self.make_single_operator_list(self.occupied_2)

    def set_up_list(self):
        self.up_list = self.make_single_operator_list(self.up)

    def set_up10_list(self):
        self.up10_list = self.make_single_operator_list(self.up10)

    def set_up21_list(self):
        self.up21_list = self.make_single_operator_list(self.up21)

    def set_up20_list(self):
        self.up20_list = self.make_single_operator_list(self.up20)

    def set_down01_list(self):
        self.down01_list = self.make_single_operator_list(self.down01)

    def set_down12_list(self):
        self.down12_list = self.make_single_operator_list(self.down12)

    def set_down02_list(self):
        self.down02_list = self.make_single_operator_list(self.down02)

    def set_down_list(self):
        self.down_list = self.make_single_operator_list(self.down)

    def set_empty_list(self):
        self.empty_list = self.make_single_operator_list(self.empty)

    def set_exchange_list(self):
        self.exchange_list = self.make_multi_operator_list([self.up, self.down])  # a_dagger*a

    def set_exchange_list_G(self):
        self.exchange_list_G = self.make_multi_operator_list([self.up21, self.down])  # c_dagger * a

    def set_exchange_list_GG(self):
        self.exchange_list_GG = self.make_multi_operator_list([self.down, self.up21])  #

    def set_exchange_list_C(self):
        self.exchange_list_C = self.make_multi_operator_list([self.up21, self.down12])  # c_dagger * c

    def set_mono_H_list(self):
        self.mono_list = self.make_single_operator_list(self.H_i)

    ### Tools for moving back and forth between full Hamiltonian and manifold(s)

    def electronic_vector_of_ones_kron(self, position, item):
        n = self.num_sites
        ones_list = [np.ones(self.N) for i in range(n - 1)]
        ones_list.insert(position, item)
        vec = ones_list.pop(0)
        for next_item in ones_list:
            vec = np.kron(vec, next_item)
        return vec

    def set_electronic_total_occupation_number(self):
        n = self.num_sites
        single_mode_occ = np.arange(self.N)
        occ_num = self.electronic_vector_of_ones_kron(0, single_mode_occ)
        for i in range(1, n):
            occ_num += self.electronic_vector_of_ones_kron(i, single_mode_occ)
        self.H_dim = occ_num.size
        self.electronic_total_occupation_number = occ_num

    def electronic_manifold_mask(self, manifold_num):
        """Creates a boolean mask to describe which states obey the truncation
           size collectively
"""
        manifold_inds = np.where(self.electronic_total_occupation_number == manifold_num)[0]
        return manifold_inds

    def electronic_subspace_mask(self, min_occ_num, max_occ_num):
        """Creates a boolean mask to describe which states obey the range of
            electronic occupation collectively
"""
        manifold_inds = np.where((self.electronic_total_occupation_number >= min_occ_num) &
                                 (self.electronic_total_occupation_number <= max_occ_num))[0]
        return manifold_inds

    def extract_coherence(self, O, manifold1, manifold2):
        """Returns result of projecting the Operator O onto manifold1
            on the left and manifold2 on the right
"""
        manifold1_inds = self.electronic_manifold_mask(manifold1)
        manifold2_inds = self.electronic_manifold_mask(manifold2)
        O = O[manifold1_inds, :]
        O = O[:, manifold2_inds]
        return O

    def extract_manifold(self, O, manifold_num):
        """Projects operator into the given electronic excitation manifold
"""
        return self.extract_coherence(O, manifold_num, manifold_num)

    def coherence_to_full(self, O, manifold1, manifold2):
        """Creates an array of zeros of the size of the full Hilbert space,
            and fills the correct entries with the operator O existing in
            a particular optical coherence between manifolds
"""
        Ofull = np.zeros(self.electronic_hamiltonian.shape, dtype=O.dtype)
        manifold1_inds = self.electronic_manifold_mask(manifold1)
        manifold2_inds = self.electronic_manifold_mask(manifold2)
        for i in range(manifold2_inds.size):
            ind = manifold2_inds[i]
            Ofull[manifold1_inds, ind] = O[:, i]
        return Ofull

    def manifold_to_full(self, O, manifold_num):
        """Creates an array of zeros of the size of the full Hilbert space,
            and fills the correct entries with the operator O existing in
            a single optical manifold
"""
        return self.coherence_to_full(O, manifold_num, manifold_num)

    def extract_electronic_subspace(self, O, min_occ_num, max_occ_num):
        """Projects operator into the given electronic excitation manifold(s)
"""
        manifold_inds = self.electronic_subspace_mask(min_occ_num, max_occ_num)
        O = O[manifold_inds, :]
        O = O[:, manifold_inds]
        return O

    def truncate_operator(self, O, pdc):
        """ Truncates operator O to the subspace of the maximum reachable manifold ,
        given the phase-discrimination condition
        """
        if isinstance(pdc[0], tuple):
            pdc_list = []
            for i in pdc:
                pdc_list += list(i)

            spectroscopy_order = sum(pdc_list)

        else:
            spectroscopy_order = sum(pdc)

        if (spectroscopy_order)%2 == 1:
            maximum_manifold = (spectroscopy_order +1)/2
        else:
            maximum_manifold = (spectroscopy_order)/2

        self.truncated_O = self.extract_electronic_subspace(O,0,maximum_manifold)

        return self.truncated_O

    ### Tools for making the Hamiltonian

    def make_electronic_hamiltonian(self):
        if self.N == 2:
            ham = self.energies[0] * self.occupied_list[0]
            for i in range(1, self.num_sites):
                ham += self.energies[i] * self.occupied_list[i]

            for i in range(len(self.exchange_list)):
                ham += self.couplings[i] * self.exchange_list[i]
                ham += np.conjugate(self.couplings[i]) * self.exchange_list[i].T

            return ham
        if self.N == 3:

            self.H_i = []
            self.H1 = np.zeros((self.H_dim,self.H_dim))
            self.H2 = np.zeros((self.H_dim,self.H_dim))
            self.H3 = np.zeros((self.H_dim,self.H_dim))

            for i in range(self.num_sites):
                self.mono_H = np.zeros((self.N, self.N))
                self.mono_H[1][1] = self.energies[i]
                self.mono_H[2][2] = self.double_energies[i]
                self.H_i.append(self.mono_H)

            self.HI = self.make_HI_i(self.H_i)

            self.H0 = np.zeros((self.H_dim,self.H_dim))

            for i in range(len(self.HI)):
                self.H0 += self.HI[i]

            for i in range(len(self.exchange_list)):
                self.H1 += self.J_list[i] * self.exchange_list[i]
                self.H1 += np.conjugate(self.J_list[i]) * self.exchange_list[i].T

                self.H2 += self.K_list[i] * (self.exchange_list_G[i] + self.exchange_list_GG[i].T)
                self.H2 += np.conjugate(self.K_list[i]) * (self.exchange_list_G[i].T + self.exchange_list_GG[i])

                self.H3 += self.L_list[i] * self.exchange_list_C[i]
                self.H3 += np.conjugate(self.L_list[i]) * self.exchange_list_C[i].T

            electronic_hamiltonian = self.H0 + self.H1 + self.H2 + self.H3

            return electronic_hamiltonian

    def set_electronic_hamiltonian(self):
        self.electronic_hamiltonian = self.make_electronic_hamiltonian()
        return self.electronic_hamiltonian

    def get_electronic_hamiltonian(self, *, manifold_num='all'):
        if manifold_num == 'all':
            return self.make_electronic_hamiltonian()
        else:
            return self.extract_manifold(self.electronic_hamiltonian, manifold_num)

    ### Tools for diagonalizing the Hamiltonian

    def make_manifold_eigensystem(self, manifold_num):
        h = self.get_electronic_hamiltonian(manifold_num=manifold_num)
        e, v = np.linalg.eigh(h)
        sort_inds = e.argsort()
        e = e[sort_inds]
        v = v[:, sort_inds]
        return e, v

    def set_manifold_eigensystems(self):
        self.electronic_eigenvalues_by_manifold = []
        self.electronic_eigenvectors_by_manifold = []
        for i in range(self.num_sites + 1):
            e, v = self.make_manifold_eigensystem(i)
            self.electronic_eigenvalues_by_manifold.append(e)
            self.electronic_eigenvectors_by_manifold.append(v)

    def get_eigensystem_by_manifold(self, manifold_num):
        e = self.electronic_eigenvalues_by_manifold[manifold_num]
        v = self.electronic_eigenvectors_by_manifold[manifold_num]
        return e, v

    def set_electronic_eigensystem(self):
        H = self.electronic_hamiltonian
        eigvecs = np.zeros(H.shape)
        d = np.zeros(H.shape)
        for i in range(self.num_sites + 1):
            e, v = self.get_eigensystem_by_manifold(i)
            if i == 1:
                self.exciton_energies = e
            eigvecs += self.manifold_to_full(v, i)
            d += self.manifold_to_full(np.diag(e), i)
        Hd = eigvecs.T.dot(H.dot(eigvecs))
        if np.allclose(Hd, d):
            pass
        else:
            raise Exception('Diagonalization by manifold failed')

        self.electronic_eigenvectors = eigvecs
        self.electronic_eigenvalues = d.diagonal()

    ### Tools for making the dipole operator

    def make_mu_site_basis(self, pol):
        if self.N == 2:
            pol_dict = {'x': 0, 'y': 1, 'z': 2}
            d = self.dipoles[:, pol_dict[pol]]
            self.mu = d[0] * (self.up_list[0] + self.down_list[0])
            for i in range(1, len(self.up_list)):
                self.mu += d[i] * (self.up_list[i] + self.down_list[i])

        elif self.N == 3:
            pol_dict = {'x': 0, 'y': 1, 'z': 2}
            d = self.dipoles[..., pol_dict[pol]]
            self.mu_10_01 = d[0,0] * (self.up10_list[0].copy() + self.down01_list[0].copy())
            self.mu_21_12 = d[0,1] * (self.up21_list[0].copy() + self.down12_list[0].copy())
            self.mu_20_02 = d[0,2] * (self.up20_list[0].copy() + self.down02_list[0].copy())
            for i in range(1, len(self.up10_list)):
                self.mu_10_01 += d[i,0] * (self.up10_list[i] + self.down01_list[i])
                self.mu_21_12 += d[i,1] * (self.up21_list[i] + self.down12_list[i])
                self.mu_20_02 += d[i,2] * (self.up20_list[i] + self.down02_list[i])
            self.mu = self.mu_10_01 + self.mu_21_12 + self.mu_20_02

    def make_mu_dict_site_basis(self):
        if self.N == 2:
            self.mu_dict = dict()
            for pol in self.pols:
                self.make_mu_site_basis(pol)
                self.mu_dict[pol] = self.mu.copy()

        elif self.N == 3 :
            self.mu_dict = dict()
            for pol in self.pols:
                self.make_mu_site_basis(pol)
                self.mu_dict[pol] = self.mu.copy()

    def make_mu_up_site_basis(self, pol):
        if self.N == 2:
            pol_dict = {'x': 0, 'y': 1, 'z': 2}
            d = self.dipoles[:, pol_dict[pol]]
            self.mu_ket_up = self.up_list[0].copy() * d[0]
            for i in range(1, len(self.up_list)):
                self.mu_ket_up += self.up_list[i] * d[i]

        elif self.N == 3:
            pol_dict = {'x': 0, 'y': 1, 'z': 2}
            d = self.dipoles[..., pol_dict[pol]]
            self.mu_ket_up10 = self.up10_list[0].copy() * d[0, 0]
            self.mu_ket_up21 = self.up21_list[0].copy() * d[0, 1]
            self.mu_ket_up20 = self.up20_list[0].copy() * d[0, 2]
            for i in range(1, len(self.up10_list)):
                self.mu_ket_up10 += self.up10_list[i] * d[i, 0]
                self.mu_ket_up21 += self.up21_list[i] * d[i, 1]
                self.mu_ket_up20 += self.up20_list[i] * d[i, 2]
            self.mu_ket_up = self.mu_ket_up10 + self.mu_ket_up21 + self.mu_ket_up20

    def make_mu_up_dict_site_basis(self):
        if self.N == 2:
            self.mu_up_dict = dict()
            for pol in self.pols:
                self.make_mu_up_site_basis(pol)
                self.mu_up_dict[pol] = self.mu_ket_up.copy()
        if self.N == 3:
            self.mu_up_dict = dict()
            for pol in self.pols:
                self.make_mu_up_site_basis(pol)
                self.mu_up_dict[pol] = self.mu_ket_up.copy()


    def make_mu_up_3d(self):
        self.mu_up_3d = np.zeros((self.H_dim,self.H_dim,3))
        self.mu_up_3d[:, :, 0] = self.mu_up_dict['x']
        self.mu_up_3d[:, :, 1] = self.mu_up_dict['y']
        self.mu_up_3d[:, :, 2] = self.mu_up_dict['z']

class PolymerVibrations():
    def __init__(self,yaml_file,*,mask_by_occupation_num=True,
                 separable_manifolds = 'auto'):
        """Initial set-up involves unpacking the vibrational_frequencies, 
            which must be passed as a nested list. Each site may have N 
            vibrational modes, and each has a frequency, a displacement
            and a frequency shift for the excited state for sites a, b, ...
"""
        with open(yaml_file) as yamlstream:
            params = yaml.load(yamlstream,Loader=yaml.SafeLoader)
        self.base_path = os.path.split(yaml_file)[0]

        self.save_path = os.path.join(self.base_path,'closed')
        os.makedirs(self.save_path,exist_ok=True)

        try:
            max_manifold = params['maximum_manifold']
        except:
            max_manifold = 'auto'

        try:
            self.RWA = params['RWA']
        except KeyError:
            self.RWA = True

        self.Polymer = Polymer(params['site_energies'],params['site_couplings'],
                               np.array(params['dipoles']),maximum_manifold=max_manifold)

        self.maximum_manifold = self.Polymer.maximum_manifold
        
        self.truncation_size = params['initial truncation size']

        if separable_manifolds == 'auto':
            try:
                int_conv_bath = params['bath']['site_internal_conversion_bath']
                self.separable_manifolds = False
            except KeyError:
                try:
                    opt_rel = params['bath']['optical_relaxation_gamma']
                    self.separable_manifolds = False
                except KeyError:
                    self.separable_manifolds = True

        else:
            self.separable_manifolds = separable_manifolds
        
        self.params = params
        
        self.occupation_num_mask = mask_by_occupation_num
        self.set_vibrations()
        self.set_vibrational_ladder_operators()
        
        e_ham = self.Polymer.extract_electronic_subspace(self.Polymer.electronic_hamiltonian,0,self.maximum_manifold)
            
        self.total_hamiltonian = np.kron(e_ham,self.vibrational_identity)
        self.add_vibrations()

        self.set_H()
        self.save_H()
        
        self.make_condon_mu()

        self.save_mu()

    def set_H(self):
        self.H = {}
        if self.separable_manifolds:
            for i in range(self.maximum_manifold+1):
                H = self.extract_vibronic_manifold(self.total_hamiltonian,i)
                self.H[str(i)] = H
        else:
            self.H['all_manifolds'] = self.total_hamiltonian

    def save_H(self):
        np.savez(os.path.join(self.save_path,'H.npz'),**self.H)

    def vibrational_occupation_to_indices(self,vibration,occ_num,
                                          manifold_num):
        single_mode_occ = np.arange(self.truncation_size)
        vib_occ = self.vibrational_vector_of_ones_kron(vibration,
                                                       single_mode_occ)
        masked_single_mode_occ = vib_occ[self.vibrational_mask]

        electronic_manifold_hamiltonian = self.Polymer.get_electronic_hamiltonian(manifold_num = manifold_num)
        elec_size = electronic_manifold_hamiltonian.shape[0]
        
        masked_single_mode_occ = np.kron(np.ones(elec_size),
                                         masked_single_mode_occ)
        return np.where(masked_single_mode_occ == occ_num)[0]

    def electronic_occupation_to_indices(self,site_num,manifold_num):
        single_mode_occ = np.arange(2)
        elec_occ = self.Polymer.electronic_vector_of_ones_kron(site_num,
                                                       single_mode_occ)
        mask = self.Polymer.electronic_manifold_mask(manifold_num)
        masked_elec_occ = elec_occ[mask]
        masked_elec_occ = np.kron(masked_elec_occ,
                                  np.ones(self.vibrational_size))

        return np.where(masked_elec_occ == 1)[0]

    def vibronic_manifold_mask(self,manifold_num):
        """Gets the indices of the Hilbert space that occupy a particular 
            electronic manifold, including all vibrational degrees of 
            freedom from that manifold
"""
        vib_size = self.vibrational_size
        vib_ones = np.ones(vib_size,dtype='int')
        elec_occ = self.Polymer.electronic_total_occupation_number
        elec_inds = np.where(elec_occ <= self.maximum_manifold)[0]
        elec_occ = elec_occ[elec_inds]
        vibronic_occupation_number = np.kron(elec_occ,vib_ones)
        manifold_inds = np.where(vibronic_occupation_number == manifold_num)[0]
        return manifold_inds

    def extract_vibronic_coherence(self,O,manifold1,manifold2):
        """Returns result of projecting the Operator O onto manifold1
            on the left and manifold2 on the right
"""
        manifold1_inds = self.vibronic_manifold_mask(manifold1)
        manifold2_inds = self.vibronic_manifold_mask(manifold2)
        O = O[manifold1_inds,:]
        O = O[:,manifold2_inds]
        return O
    
    def extract_vibronic_manifold(self,O,manifold_num):
        """Projects operator into the given electronic excitation manifold
"""
        return self.extract_vibronic_coherence(O,manifold_num,manifold_num)
    
    def add_vibrations(self):
        v0 = self.empty_vibrations
        v1 = self.occupied_vibrations
        if self.Polymer.N == 3:
            v2 = self.doubly_occupied_vibrations
        self.vibrational_hamiltonian = np.zeros(self.total_hamiltonian.shape)
        for i in range(len(v0)):
            self.vibrational_hamiltonian += v0[i]
            self.vibrational_hamiltonian += v1[i]
            if self.Polymer.N == 3:
                self.vibrational_hamiltonian += v2[i]

        self.total_hamiltonian = self.total_hamiltonian + self.vibrational_hamiltonian

    # def add_linear_couplings(self):
    #     self.linear_coupling_hamiltonian = np.zeros(self.total_hamiltonian.shape)

    #     for i in range(self.num_vibrations):
    #         try:
    #             coupling_list = self.params['vibrations'][i]['linear_couplings']
    #             self.linear_coupling_hamiltonian += self.linear_site_vibrational_couplings(i,manifold_num)
    #         except KeyError:
    #             pass
    #     self.total_hamiltonian = self.total_hamiltonian + self.linear_coupling_hamiltonian

    def set_vibrations(self):
        vibration_params = self.params['vibrations']
        # Vibrations in the ground manifold are assumed to be diagonal
        
        
        emp_vibs = [self.construct_vibrational_hamiltonian(mode_dict,0)
                    for mode_dict in vibration_params]
        self.num_vibrations = len(emp_vibs)
        occ_vibs = [self.construct_vibrational_hamiltonian(mode_dict,1)
                    for mode_dict in vibration_params]
        if self.Polymer.N == 3:
            doub_occ_vibs = [self.construct_vibrational_hamiltonian(mode_dict,2)
                        for mode_dict in vibration_params]

        if self.occupation_num_mask:
            self.set_vibrational_total_occupation_number()
        else:
            N = self.truncation_size
            nv = self.num_vibrations
            self.vibrational_size = N**nv
            self.vibrational_mask = (np.arange(N**nv),)
            self.vibrational_identity = np.eye(N**nv)
        empty_vibrations = self.kron_up_vibrations(emp_vibs)
        occupied_vibrations = self.kron_up_vibrations(occ_vibs)
        if self.Polymer.N == 3:
            doubly_occupied_vibrations = self.kron_up_vibrations(doub_occ_vibs)

        self.empty_vibrations = []
        self.occupied_vibrations = []
        if self.Polymer.N == 3:
            self.doubly_occupied_vibrations = []
        
        for i in range(self.num_vibrations):
            site_index = vibration_params[i]['site_label']
            empty = self.Polymer.empty_list[site_index]
            empty = self.Polymer.extract_electronic_subspace(empty,0,self.maximum_manifold)
            occupied = self.Polymer.occupied_list[site_index]
            occupied = self.Polymer.extract_electronic_subspace(occupied,0,self.maximum_manifold)
            if self.Polymer.N == 3:
                doubly_occupied = self.Polymer.doubly_occupied_list[site_index]
                doubly_occupied = self.Polymer.extract_electronic_subspace(doubly_occupied,0,self.maximum_manifold)
            
            self.empty_vibrations.append(np.kron(empty,empty_vibrations[i]))
            self.occupied_vibrations.append(np.kron(occupied,occupied_vibrations[i]))
            if self.Polymer.N == 3:
                self.doubly_occupied_vibrations.append(np.kron(doubly_occupied,doubly_occupied_vibrations[i]))

    def kron_up_vibrations(self,vibrations_list):
        n = self.num_vibrations
        if n == 1:
            return vibrations_list
        new_vibrations_list = []
        for i in range(n):
            new_vibration = self.vibration_identity_kron(i,vibrations_list[i])
            if self.occupation_num_mask:
                new_vibration = self.mask_vibrational_space(new_vibration)
            new_vibrations_list.append(new_vibration)
        return new_vibrations_list
            
    def mask_vibrational_space(self,O):
        inds = self.vibrational_mask
        if type(O) is np.ndarray:
            O = O[inds[0],:].copy()
            O = O[:,inds[0]].copy()
            return O
        
        if type(O) is csr_matrix:
            pass
        else:
            O = O.tocsr()
        O = O[inds[0]]
        O = O.transpose()
        O = O[inds[0]]
        O = O.transpose()
        return O

    # def linear_site_vibrational_couplings(self,mode_num):
    #     mode_dict = self.params['vibrations'][mode_num]
    #     coupling_list = mode_dict['linear_couplings']
    #     coupling_strength, site_index_pair = coupling_list[0]
    #     vib_coupling = self.single_linear_site_vibration_coupling(site_index_pair,mode_num)
    #     coupling_ham =  vib_coupling * coupling_strength
        
    #     for i in range(1,len(coupling_list)):
    #         coupling_strength, site_index_pair = coupling_list[i]
    #         vib_coupling = self.single_linear_site_vibration_coupling(site_index_pair,mode_num,manifold_num)
    #         coupling_ham +=  vib_coupling * coupling_strength
    #     return coupling_ham
    
    def single_linear_site_vibration_coupling(self,site_index_pair,mode_num,manifold_num):
        electronic_matrix = self.site_basis(0,manifold_num)
        electronic_matrix[:,:] = 0
        i,j = site_index_pair
        electronic_matrix[i,j] = 1
        electronic_matrix[j,i] = 1

        x = LadderOperators(self.truncation_size).x
        x = x[:self.truncation_size,:self.truncation_size]
        vib_coupling = self.vibration_identity_kron(mode_num,x,manifold_num)
        return kron(electronic_matrix,vib_coupling)

    def vibration_identity_kron(self,position,item):
        """Takes in a single vibrational hamiltonians and krons it with the correct 
            number of vibrational identities, inserting it into its position as indexed by its mode
            position as specified in the input file"""
        identities = [np.eye(self.truncation_size) for n in
                      range(self.num_vibrations-1)]
        identities.insert(position,item)
        mat = identities.pop(0)
        for next_item in identities:
            mat = np.kron(mat,next_item)
        return mat

    def vibrational_vector_of_ones_kron(self,position,item):
        """Takes in a single vibrational hamiltonians and krons it with the correct 
            number of vibrational identities, inserting it into its position as indexed by its mode
            position as specified in the input file"""
        N = self.truncation_size
        nv = self.num_vibrations
        ones_list = [np.ones(N) for i in range(nv-1)]
        ones_list.insert(position,item)
        vec = ones_list.pop(0)
        for next_item in ones_list:
            vec = np.kron(vec,next_item)
        return vec

    def set_vibrational_total_occupation_number(self):
        N = self.truncation_size
        nv = self.num_vibrations
        if nv == 0:
            single_mode_occ = np.array([0])
        else:
            single_mode_occ = np.arange(N)
        occ_num = self.vibrational_vector_of_ones_kron(0,single_mode_occ)
        for i in range(1,nv):
            occ_num += self.vibrational_vector_of_ones_kron(i,single_mode_occ)
        self.vibrational_total_occupation_number = occ_num
        self.vibrational_mask = np.where(occ_num < N)
        self.vibrational_size = self.vibrational_mask[0].size
        self.vibrational_identity = np.eye(self.vibrational_size)

    def construct_vibrational_hamiltonian(self,single_mode,electronic_occupation):
        """For each vibrational mode, construct a list of sparse matrices 
            defining the vibrational hamiltonian for that mode in each 
            excited state"""
        w = single_mode['omega_g']
        lam = single_mode['reorganization'][electronic_occupation]
        d  = single_mode['displacement'][electronic_occupation]
        kin = single_mode['kinetic'][electronic_occupation]
        pot = single_mode['potential'][electronic_occupation]
        aho = DisplacedAnharmonicOscillator(self.truncation_size)
        aho.set_ham(lam,d,kin,pot)
        return 0.5 * w * aho.ham

    def construct_vibrational_ladder_operator(self,single_mode,electronic_occupation):
        """Construct ladder operator given the electronic occupation for 
            that site"""
        w = single_mode['omega_g']
        d  = single_mode['displacement'][electronic_occupation]
        L = LadderOperators(self.truncation_size,disp=d,extra_size=1)
        return L.ad

    def set_vibrational_ladder_operators(self):
        vibration_params = self.params['vibrations']
        emp_ups = []
        occ_ups = []

        for i in range(len(vibration_params)):
            ad = self.construct_vibrational_ladder_operator(vibration_params[i],0)
            emp_ups.append(ad)

            ad = self.construct_vibrational_ladder_operator(vibration_params[i],1)
            occ_ups.append(ad)

        empty_ups = self.kron_up_vibrations(emp_ups)
        occupied_ups = self.kron_up_vibrations(occ_ups)
        self.empty_ups = []
        self.occupied_ups = []
        self.total_ups = []
        
        for i in range(self.num_vibrations):
            site_index = vibration_params[i]['site_label']
            empty = self.Polymer.empty_list[site_index]
            empty = self.Polymer.extract_electronic_subspace(empty,0,self.maximum_manifold)
            occupied = self.Polymer.occupied_list[site_index]
            occupied = self.Polymer.extract_electronic_subspace(occupied,0,self.maximum_manifold)
            
            self.empty_ups.append(np.kron(empty,empty_ups[i]))
            self.occupied_ups.append(np.kron(occupied,occupied_ups[i]))

            self.total_ups.append(self.empty_ups[i] + self.occupied_ups[i])

    def make_condon_mu(self):
        H_shape = self.total_hamiltonian.shape
        mu_shape = (*H_shape,3)
        # self.mu = np.zeros(mu_shape)
        self.mu_ket_up = np.zeros(mu_shape)
        for i in range(3):
            pol = self.Polymer.pols[i]
            # elec_mu = self.Polymer.mu_dict[pol]
            # mu = self.Polymer.extract_electronic_subspace(elec_mu,0,self.maximum_manifold)
            # self.mu[:,:,i] = np.kron(mu,self.vibrational_identity)

            elec_mu_up = self.Polymer.mu_up_dict[pol]
            mu_up = self.Polymer.extract_electronic_subspace(elec_mu_up,0,self.maximum_manifold)
            if not self.RWA:
                if self.separable_manifolds:
                    raise Exception('manifolds cannot be separated when moving beyond the RWA')
                mu_up += mu_up.T
            
            self.mu_ket_up[:,:,i] = np.kron(mu_up,self.vibrational_identity)

    def save_mu_by_manifold(self):
        mu_dict = {}
        for i in range(self.maximum_manifold):
            j = i+1
            mu = self.extract_vibronic_coherence(self.mu_ket_up[:,:,0],j,i)
            mu_shape = (*mu.shape,3)
            mu3d = np.zeros(mu_shape)
            mu3d[:,:,0] = mu[:,:]
            mu3d[:,:,1] = self.extract_vibronic_coherence(self.mu_ket_up[:,:,1],j,i)
            mu3d[:,:,2] = self.extract_vibronic_coherence(self.mu_ket_up[:,:,2],j,i)
            key = str(i) + '_to_' + str(j)
            mu_dict[key] = mu3d
        np.savez(os.path.join(self.save_path,'mu_original_H_basis.npz'),**mu_dict)

    def save_mu_all_manifolds(self):
        mu_up = self.mu_ket_up
        mu_dict = {'up':mu_up}

        np.savez(os.path.join(self.save_path,'mu_original_H_basis.npz'),**mu_dict)

    def save_mu(self):
        if self.separable_manifolds:
            self.save_mu_by_manifold()
        else:
            self.save_mu_all_manifolds()
        

class DiagonalizeHamiltonian:

    def __init__(self,file_path,*,qr_flag=False,check_eigenvectors = True):
        self.base_path = file_path
        self.load_path = os.path.join(file_path,'closed')
        self.save_path = self.load_path
        self.qr_flag = qr_flag
        self.r_mats = {}
        self.check_eigenvectors = check_eigenvectors

        self.load_H()
        self.load_mu()

        self.eigenvalues = {}
        self.eigenvectors = {}
        self.H_diagonalization_time = {}

        self.mu_eigen_basis = {}

        self.diagonalize()
        self.transform_mu()
        self.save_eigsystem()
        self.save_mu()

    def load_H(self):
        H_save_name = os.path.join(self.load_path,'H.npz')
        with np.load(H_save_name) as H_archive:
            self.manifolds = list(H_archive.keys())
            self.H = {key:H_archive[key] for key in self.manifolds}

    def load_mu(self):
        mu_save_name = os.path.join(self.load_path,'mu_original_H_basis.npz')
        with np.load(mu_save_name) as mu_archive:
            self.mu_keys = list(mu_archive.keys())
            self.mu = {key:mu_archive[key] for key in self.mu_keys}

    def save_timing(self):
        np.savez(os.path.join(self.save_path,'diagonalization_timings.npz')
                 ,**self.H_diagonalization_time)

    def eig(self,manifold_key):
        """Diagonalzie the requested manifold

        Args:
            manifold_key (str): options: 'all_manifolds' or '0', '1', ...

        Returns:
            tuple: (e, v) where e are the eigenvalues and v is a square array
                of eigenvectors arranged in columns
"""
        ham = self.H[manifold_key]
        eigvals, eigvecs = eigh(ham)
        
        if self.qr_flag:
            # Force degenerate eigenvectors to be orthogonal
            eigvecs, r = np.linalg.qr(eigvecs,mode='reduced')
            
        if self.check_eigenvectors:
            HV = ham.dot(eigvecs)
            D = eigvecs.T.dot(HV)
            if np.allclose(D,np.diag(eigvals),rtol=1E-11,atol=1E-11):
                pass
            else:
                warnings.warn('Using eigenvectors to diagonalize hamiltonian does not result in the expected diagonal matrix to tolerance, largest deviation is {}'.format(np.max(np.abs(D - np.diag(eigvals)))))
        
        sort_indices = eigvals.argsort()
        eigvals.sort()
        eigvecs = eigvecs[:,sort_indices]
        if self.qr_flag:
            r = r[:,sort_indices]
            self.r_mats[manifold_key] = r
        # I choose to pick the phase of my eigenvectors such that the state
        # which has the largest magnitude overlap is positive. For
        # sufficiently small d, and alpha close to 1, this will be the
        # overlap between the same excited and ground states.
        for i in range(eigvals.size):
            max_index = np.argmax(np.abs(eigvecs[:,i]))
            if eigvecs[max_index,i] < 0:
                eigvecs[:,i] *= -1

        return eigvals, eigvecs

    def diagonalize(self):
        for manifold in self.manifolds:
            t0 = time.time()
            e, v = self.eig(manifold)
            t1 = time.time()
            self.H_diagonalization_time[manifold] = t1-t0
            self.eigenvalues[manifold] = e
            self.eigenvectors[manifold] = v
        return None

    def mu_key_to_manifold_keys(self,key):
        if self.manifolds[0] == 'all_manifolds':
            starting_key = 'all_manifolds'
            ending_key = 'all_manifolds'
        else:
            starting_key, ending_key = key.split('_to_')
        return starting_key, ending_key

    def transform_mu(self):
        for key in self.mu_keys:
            old_key, new_key = self.mu_key_to_manifold_keys(key)
            
            v_old = self.eigenvectors[old_key]
            v_new = self.eigenvectors[new_key]
            mu_site_basis = self.mu[key]
            mu_t = np.zeros(mu_site_basis.shape)
            for i in range(3):
                mu_t[...,i] = v_new.T.dot(mu_site_basis[...,i]).dot(v_old)
            self.mu_eigen_basis[key] = mu_t
        return None

    def save_eigsystem(self):
        e_save_name = os.path.join(self.save_path,'eigenvalues.npz')
        v_save_name = os.path.join(self.save_path,'eigenvectors.npz')
        np.savez(e_save_name,**self.eigenvalues)
        np.savez(v_save_name,**self.eigenvectors)
        return None

    def save_mu(self):
        mu_save_name = os.path.join(self.save_path,'mu.npz')
        np.savez(mu_save_name,**self.mu_eigen_basis)
        return None


def SHO_basis(n,xvalues):
    coef = [0 for i in range(n)]
    coef.append(1)
    norm = 1/(np.pi**(1/4)*2**(n/2) * np.sqrt(factorial(n)))
    return  norm * np.exp(-xvalues**2/2) * hermval(xvalues,coef)

def oscillator_1d(array_1d,x):
    ans = np.zeros(x.shape,dtype=array_1d.dtype)
    for i in range(len(array_1d)):
        ans += SHO_basis(i,x) * array_1d[i]
    return ans



#Standard python libraries
import os
import warnings
import copy
import time

#Dependencies - numpy, scipy, matplotlib
import numpy as np
import numpy.polynomial.chebyshev as npch

from ufss import CompositeDiagrams
from ufss.perturbative_calculations import UF2BaseClass
from ufss.perturbative_calculations import ClosedBaseClass
from ufss.perturbative_calculations import perturbative_container
from ufss.perturbative_calculations import ChebPoly, cheb_perturbative_container

class UF2ClosedEngine(ClosedBaseClass,UF2BaseClass):
    """This class is designed to calculate perturbative wavepackets in the
        light-matter interaction given the eigenvalues of the unperturbed 
        hamiltonian and the material dipole operator evaluated in the
        eigenbasis of the unperturbed hamiltonian.

    Args:
        file_path (string): path to folder containing eigenvalues and the
            dipole operator for the system Hamiltonian
        initial_state (int): index of initial state for psi^0

"""
    def __init__(self,file_path,*,initial_state=0,
                 detection_type = 'polarization'):
        UF2BaseClass.__init__(self,file_path,detection_type = detection_type)
        ClosedBaseClass.__init__(self,detection_type = detection_type)
        self.slicing_time = 0
        self.interpolation_time = 0
        self.expectation_time = 0
        self.next_order_expectation_time = 0
        self.next_order_time = 0
        self.get_psi_time = 0
        self.convolution_time = 0
        self.extend_time = 0
        self.mask_time = 0
        self.dipole_time = 0
        self.automation_time = 0
        self.diagram_to_signal_time = 0
        self.efield_mask_time = 0
        self.method = 'UF2'

        self.initial_state = initial_state

        self.load_eigensystem()

        self.set_min_max_manifolds()

        self.load_mu()

    def add_impulsive_psis(self,a,b):
        """Add two perturbative_containers together
"""
        manifold_key = a.manifold_key
        pdc = a.pdc
        sim_a = a.simultaneous
        sim_b = b.simultaneous
        if sim_a != sim_b:
            raise Exception('Cannot add impulsive perturbative containers if they do not have the same number of simultaneous interactions')
        else:
            simultaneous = sim_a
        if np.allclose(a.t0,b.t0):
            t = a.t[1:2]
            f_a = a.f[:,1]
            f_b = b.f[:,1]
            pulse_number = a.pulse_number
            t0 = a.t0
        else:
            raise Exception('Cannot add two impulsive perturbative containers unless they have the same value of t0')

        f = np.zeros((a.bool_mask.size,t.size),dtype='complex')
        f[a.bool_mask,:] = f_a[:,np.newaxis]
        f[b.bool_mask,:] += f_b[:,np.newaxis]
        f *= 2 * simultaneous

        bool_mask = np.logical_or(a.bool_mask,b.bool_mask)
        f = f[bool_mask,:]
        
        ab=perturbative_container(t,f,bool_mask,pulse_number,manifold_key,
                                   pdc,t0,simultaneous = simultaneous)
        return ab

    def add_psis(self,a,b):
        """Add two perturbative_container objects together
"""
        if not np.allclose(a.pdc,b.pdc):
            raise Exception('Cannot add density matrices with different phase-discrimination conditions')
        if 'all_manifolds' in self.manifolds:
            pass
        else:
            if a.manifold_key != b.manifold_key:
                raise Exception('Cannot add rho_container objects that exist in different manifolds')
            
        if a.impulsive:
            return self.add_impulsive_psis(a,b)
        
        manifold_key = a.manifold_key
        
        pdc = a.pdc

        if self.method == 'UF2':
            tmin = min(a.t[0],b.t[0])
            tmax = max(a.t[-1],b.t[-1])
            dt = min(a.t[1]-a.t[0],b.t[1]-b.t[0])
            t = np.arange(tmin,tmax+dt*0.9,dt)
        elif self.method == 'chebyshev':
            l1,u1 = a.dom
            l2,u2 = b.dom
            dom = np.array([min(l1,l2),max(u1,u2)])
            midpoint = (dom[1] + dom[0])/2
            halfwidth = (dom[1] - dom[0])/2
            order = max(a.order,b.order)
            t = npch.chebpts1(order) * halfwidth + midpoint
            
        f = np.zeros((a.bool_mask.size,t.size),dtype='complex')
        f_a = a(t)
        f_b = b(t)

        if a.t0 > b.t0:
            t0 = a.t0
            pulse_number = a.pulse_number
        else:
            t0 = b.t0
            pulse_number = b.pulse_number

        # try:
        #     eva = self.eigenvalues['all_manifolds'][a.bool_mask]
        #     evb = self.eigenvalues['all_manifolds'][b.bool_mask]
        # except KeyError:
        #     eva = self.eigenvalues[a.manifold_key][a.bool_mask]
        #     evb = self.eigenvalues[b.manifold_key][b.bool_mask]

        # if self.interaction_picture_shift:
        #     t_exp_a = t0 - a.t0
        #     t_exp_b = t0 - b.t0
        #     f_a *= np.exp(eva[:,np.newaxis]*t_exp_a)
        #     f_b *= np.exp(evb[:,np.newaxis]*t_exp_b)
        # else:
        #     pass

        f[a.bool_mask,:] = f_a
        f[b.bool_mask,:] += f_b

        bool_mask = np.logical_or(a.bool_mask,b.bool_mask)
        f = f[bool_mask,:]

        if self.method == 'UF2':
            ab=perturbative_container(t,f,bool_mask,pulse_number,
                                      manifold_key,pdc,t0)
        elif self.method == 'chebyshev':
            ab=cheb_perturbative_container(t,f,bool_mask,pulse_number,
                                           manifold_key,pdc,t0,dom=dom)
        return ab
        
    def set_efields(self,times_list,efields_list,centers_list,
                    phase_discrimination,*,reset_calculations = True,
                    plot_fields = False):
        
        self.interaction_picture_calculations = True
        
        UF2BaseClass.set_efields(self,times_list,efields_list,centers_list,
                                 phase_discrimination,
                                 reset_calculations = reset_calculations,
                                 plot_fields = plot_fields)
        
        # Initialize unperturbed wavefunction
        self.set_psi0(self.initial_state)

    def get_psi_eigen_basis(self,t,key):
        psi_obj = self.psis[key]
        mask = psi_obj.bool_mask
        all_e = self.eigenvalues[psi_obj.manifold_key]
        e = all_e[mask]
        psi = psi_obj(t)*np.exp(-1j*e[:,np.newaxis]*t[np.newaxis,:])
        full_size = all_e.size
        total_psi = np.zeros(full_size,t.size)
        total_psi[mask,:] = psi
        return total_psi

    def get_psi_site_basis(self,t,key):
        psi_obj = self.psis[key]
        mask = psi_obj.bool_mask
        manifold_num = self.manifold_key_to_number(psi_obj.manifold_key)
        e = self.eigenvalues[manifold_num][mask]
        psi = psi_obj(t)*np.exp(-1j*e[:,np.newaxis]*t[np.newaxis,:])
        ev = self.eigenvectors[manifold_num][:,mask]
        new_psi = ev.dot(psi)
        full_size = 0
        manifold_sizes = []
        for i in range(len(self.eigenvalues)):
            manifold_sizes.append(self.eigenvectors[manifold_num].shape[0])
            full_size += manifold_sizes[-1]
        total_psi = np.zeros(full_size,t.size)
        start = 0
        for i in range(len(self.eigenvalues)):
            if i == manifold_num:
                end = start + manifold_sizes[i]
                total_psi[start:end,:] = new_psi
            else:
                start += manifold_sizes[i]
        return total_psi

    def get_psi_site_basis_by_order(self,t,order):
        keys = self.psis.keys()
        order_keys = []
        for key in keys:
            if len(key) == 2*order:
                order_keys.append(key)
        psi_total = self.get_psi_site_basis(t,order_keys.pop(0))
        for key in order_keys:
            psi_total += self.get_psi_site_basis(t,key)

        return psi_total

    def set_psi0(self,initial_state):
        """Creates the unperturbed wavefunction. This code does not 
            support initial states that are coherent super-positions of 
            eigenstates. To perform thermal averaging, recalculate spectra 
            for each initial state that contributes to the thermal ensemble.
        Args:
            initial_state (int): index for initial eigenstate in GSM
"""
        # initial state must be interpreted given the fact that masking may have been done
        try:
            trimmed_indices = np.where(self.trimming_masks[0])[0]
            initial_state = np.where(trimmed_indices == initial_state)[0]
        except AttributeError:
            pass

        t = self.efield_times[0] # can be anything of the correct length

        key = self.ordered_manifolds[0]
        psi0 = np.ones((1,t.size),dtype=complex)
        bool_mask = np.zeros(self.eigenvalues[key].size,dtype='bool')
        bool_mask[initial_state] = True

        pdc = np.zeros(self.pdc.shape,dtype=self.pdc.dtype)

        t0 = 0

        if self.method == 'UF2':
            self.psi0 = perturbative_container(t,psi0,bool_mask,None,key,pdc,
                                               t0,interp_kind='zero',
                                               interp_left_fill=1)
        elif self.method == 'chebyshev':
            self.psi0 = cheb_perturbative_container(t,psi0,bool_mask,None,
                                                    key,pdc,t0,
                                                    interp_left_fill=1,
                                                    dom = self.doms[0])

    def load_eigensystem(self):
        """Load in known eigenvalues. Must be stored as a numpy archive file,
with keys: GSM, SEM, and optionally DEM.  The eigenvalues for each manifold
must be 1d arrays, and are assumed to be ordered by increasing energy. The
energy difference between the lowest energy ground state and the lowest 
energy singly-excited state should be set to 0
"""
        eigval_save_name = os.path.join(self.base_path,'eigenvalues.npz')
        eigvec_save_name = os.path.join(self.base_path,'eigenvectors.npz')
        with np.load(eigval_save_name) as eigval_archive:
            self.manifolds = list(eigval_archive.keys())
            self.eigenvalues = {key:eigval_archive[key] for key in self.manifolds}
        with np.load(eigvec_save_name) as eigvec_archive:
            self.eigenvectors = {key:eigvec_archive[key] for key in self.manifolds}

        if '0' in self.manifolds:
            self.ordered_manifolds = [str(i) for i in range(len(self.manifolds))]
        elif 'all_manifolds' in self.manifolds:
            self.ordered_manifolds = ['all_manifolds']
        else:
            self.ordered_manifolds = ['GSM','SEM','DEM','TEM','QEM']
        
        ### store original eigenvalues for recentering purposes
        self.original_eigenvalues = copy.deepcopy(self.eigenvalues)

    def load_mu(self):
        """Load the precalculated dipole overlaps.  The dipole operator must
            be stored as a .npz file, and must contain at least one array, each with three 
            indices: (upper manifold eigenfunction, lower manifold eigenfunction, 
            cartesian coordinate).  So far this code supports up to three manifolds, and
            therefore up to two dipole operators (connecting between manifolds)"""
        file_name = os.path.join(self.base_path,'mu.npz')
        file_name_pruned = os.path.join(self.base_path,'mu_pruned.npz')
        file_name_bool = os.path.join(self.base_path,'mu_boolean.npz')

        try:
            mu_boolean_archive = np.load(file_name_bool)
            # self.mu_boolean = {'ket':mu_boolean_archive['ket'],'bra':mu_boolean_archive['bra']}
            with np.load(file_name_bool) as mu_boolean_archive:
                self.mu_boolean = {key:mu_boolean_archive[key] for key in mu_boolean_archive.keys()}
            pruned = True
            file_name = file_name_pruned
        except FileNotFoundError:
            pruned = False

        with np.load(file_name) as mu_archive:
            self.H_mu = {key:mu_archive[key] for key in mu_archive.keys()}
        if pruned == False:
            self.mu_boolean = dict()
            for key in self.H_mu.keys():
                self.mu_boolean[key] = np.ones(self.H_mu[key].shape[:2],dtype='bool')
        
    ### Tools for recursively calculating perturbed wavepackets using TDPT

    def get_H_mu(self,pulse_number,key,rotating_flag=True):
        """Calculates the dipole matrix given the electric field polarization vector"""
        overlap_matrix = ClosedBaseClass.get_H_mu(self,pulse_number,key,
                                                 rotating_flag=rotating_flag)

        boolean_matrix = np.ones(overlap_matrix.shape,dtype='bool')

        return boolean_matrix, overlap_matrix

    def electric_field_mask(self,pulse_number,key,conjugate_flag=False):
        """This method determines which molecular transitions will be 
supported by the electric field.  We assume that the electric field has
0 amplitude outside the minimum and maximum frequency immplied by the 
choice of dt and num_conv_points.  Otherwise we will inadvertently 
alias transitions onto nonzero electric field amplitudes.
"""
        ef_mask_t0 = time.time()
        if conjugate_flag:
            ending_key, starting_key = key.split('_to_')
        else:
            starting_key, ending_key = key.split('_to_')
        efield_t = self.efield_times[pulse_number]
        efield_w = self.efield_frequencies[pulse_number]
        if conjugate_flag:
            center = -self.centers[pulse_number]
        else:
            center = self.centers[pulse_number]
        try:
            eig_starting = self.eigenvalues['all_manifolds']
            eig_ending = self.eigenvalues['all_manifolds']
        except KeyError:
            eig_starting = self.eigenvalues[starting_key]
            eig_ending = self.eigenvalues[ending_key]
        
        diff = eig_ending[:,np.newaxis] - eig_starting[np.newaxis,:]
        
        if efield_t.size == 1:
            mask = np.ones(diff.shape,dtype='bool')
        else:
            # The only transitions allowed by the electric field shape are
            inds_allowed = np.where((diff - center > efield_w[0]) & (diff - center < efield_w[-1]))
            mask = np.zeros(diff.shape,dtype='bool')
            mask[inds_allowed] = 1

        self.efield_mask_time += time.time() - ef_mask_t0
        return mask

    def mask_dipole_matrix(self,boolean_matrix,overlap_matrix,
                           starting_manifold_mask,*,next_manifold_mask = None):
        """Takes as input the boolean_matrix and the overlap matrix that it 
            corresponds to. Also requires the starting manifold mask, which specifies
            which states have non-zero amplitude, given the signal tolerance requested.
            Trims off unnecessary starting elements, and ending elements. If 
            next_manifold_mask is None, then the masking is done automatically
            based upon which overlap elements are nonzero. If next_manifold_mask is
            a 1D numpy boolean array, it is used as the mask for next manifold."""
        t0 = time.time()
        if np.all(starting_manifold_mask == True):
            pass
        else:
            boolean_matrix = boolean_matrix[:,starting_manifold_mask]
            overlap_matrix = overlap_matrix[:,starting_manifold_mask]

        #Determine the nonzero elements of the new psi, in the
        #eigenenergy basis, n_nonzero
        if type(next_manifold_mask) is np.ndarray:
            n_nonzero = next_manifold_mask
        else:
            n_nonzero = np.any(boolean_matrix,axis=1)
        if np.all(n_nonzero == True):
            pass
        else:
            overlap_matrix = overlap_matrix[n_nonzero,:]

        t1 = time.time()
        self.mask_time += t1-t0

        return overlap_matrix, n_nonzero

    def mu_dot_psi(self,t,psi_in,pulse_number,up_flag=True,
                   next_manifold_mask = None):
        old_manifold_key = psi_in.manifold_key
        new_manifold_key=self.get_new_manifold_key(old_manifold_key,up_flag)
        m_nonzero = psi_in.bool_mask

        if up_flag:
            mu_key = old_manifold_key + '_to_' + new_manifold_key
        else:
            mu_key = new_manifold_key + '_to_' + old_manifold_key

        boolean_mu, mu = self.get_H_mu(pulse_number,mu_key,
                                       rotating_flag=up_flag)
            
        psi = self.get_psi(t,psi_in)

        conjugate_flag = not up_flag

        # e_mask = self.electric_field_mask(pulse_number,mu_key,
        #                                   conjugate_flag=conjugate_flag)
        # boolean_mu = boolean_mu * e_mask
        # mu = mu * e_mask

        mu, n_nonzero = self.mask_dipole_matrix(boolean_mu,mu,m_nonzero,
                                    next_manifold_mask=next_manifold_mask)

        psi = mu.dot(psi)
            
        return psi, n_nonzero

    def mu_exp_val(self,t,psis_in,pulse_number,*,mu_type = None,
                   ket_flag = None, up_flag = None):
        """Keyword arguments are provided simply to make compatible with 
            open system framework
"""
        bra_in, ket_in = psis_in
        bra_key = bra_in.manifold_key
        ket_key = ket_in.manifold_key

        bra_num = self.manifold_key_to_number(bra_key)
        ket_num = self.manifold_key_to_number(ket_key)
        
        bra_nonzero = bra_in.bool_mask
        ket_nonzero = ket_in.bool_mask

        if 'all_manifolds' in self.manifolds:
            pass
        elif np.abs(bra_num - ket_num) != 1:
            warnings.warn('Dipole only connects manifolds 0 to 1 or 1 to 2')
            return None
        t0 = time.time()
        if bra_num > ket_num:
            bra_new_mask = ket_in.bool_mask
            bra, n_mask = self.mu_dot_psi(t,bra_in,-1,up_flag=False,
                                          next_manifold_mask=bra_new_mask)
            ket = self.get_psi(t,ket_in)
        else:
            ket_new_mask = bra_in.bool_mask
            ket, n_mask = self.mu_dot_psi(t,ket_in,-1,up_flag=False,
                                          next_manifold_mask=ket_new_mask)
            bra = self.get_psi(t,bra_in)

        exp_val = np.sum(np.conjugate(bra) * ket,axis = 0)
            
        return exp_val
    
    def next_order(self,psi_in,*,up_flag=True,
                   next_manifold_mask = None,pulse_number = 0):
        """This function connects psi_p to psi_pj^(*) using a DFT convolution algorithm.

        Args:
            psi_in (psi_container): input density matrix
            pulse_number (int): index of optical pulse (0,1,2,...) can also be set to
                'impulsive'
            next_manifold_mask (np.ndarray): optional - define the states to be considered 
                in the next manifold
        
        Return:
            psi_dict (psi_container): next-order wavefunction
"""
        t_next_order0 = time.time()
        # Rotating term excites the ket
        conjugate_flag = not up_flag

        if psi_in.pulse_number == None:
            old_pulse_time = -np.inf
        else:
            old_pulse_time = self.pulse_times[psi_in.pulse_number]
        pulse_time = self.pulse_times[pulse_number]

        if np.isclose(old_pulse_time,pulse_time):
            simultaneous = psi_in.simultaneous + 1
        else:
            simultaneous = 1

        t = self.efield_times[pulse_number] + pulse_time
        dt = self.dts[pulse_number]
        old_manifold_key = psi_in.manifold_key
        new_manifold_key=self.get_new_manifold_key(old_manifold_key,up_flag)

        input_pdc = psi_in.pdc
        output_pdc = input_pdc.copy()
        if conjugate_flag:
            output_pdc[pulse_number][1] += 1
        else:
            output_pdc[pulse_number][0] += 1

        center = self.centers[pulse_number]
        efield = self.efields[pulse_number] * np.exp(-1j*center*t)

        if conjugate_flag:
            efield = np.conjugate(efield)

        # factor of 1j comes from perturbation theory
        efield = 1j * efield
        
        t0 = time.time()
        psi,n_nonzero=self.mu_dot_psi(t,psi_in,pulse_number,up_flag=up_flag,
                                     next_manifold_mask=next_manifold_mask)
        
        t1 = time.time()
        self.next_order_expectation_time += t1-t0

        psi = psi * efield[np.newaxis,:]
        
        ### UF2 specific
        try:
            ev2 = self.eigenvalues['all_manifolds']
        except KeyError:
            ev2 = self.eigenvalues[new_manifold_key]

        exp_factor_ending = np.exp(1j*ev2[n_nonzero,np.newaxis]*t[np.newaxis,:])

        psi = psi*exp_factor_ending

        t0 = time.time()

        M = self.efield_times[pulse_number].size

        convolve_fun = self.heaviside_convolve_list[pulse_number].fft_convolve2

        if M == 1:
            pass
        else:
            if self.method == 'UF2':
                psi = convolve_fun(psi,d=self.dt)
            elif self.method == 'chebyshev':
                dom = self.doms[pulse_number] + pulse_time
                chp = ChebPoly(t,psi,dom = dom)
                chp.integrate()
                psi = chp(t)

        t1 = time.time()
        self.convolution_time += t1-t0

        if self.method == 'UF2':
            psi_out = perturbative_container(t,psi,n_nonzero,pulse_number,
                                new_manifold_key,output_pdc,pulse_time,
                                simultaneous=simultaneous)
        elif self.method == 'chebyshev':
            psi_out = cheb_perturbative_container(t,psi,n_nonzero,
                                                  pulse_number,
                                                  new_manifold_key,
                                                  output_pdc,pulse_time,
                                                  simultaneous=simultaneous,
                                                  dom = dom)

        self.next_order_time += time.time() - t_next_order0
    
        return psi_out

    def get_psi(self,t,psi_obj,*,original_H_basis=False):
        t_get_psi0 = time.time()
        manifold_key = psi_obj.manifold_key
        mask = psi_obj.bool_mask
        
        try:
            e = self.eigenvalues['all_manifolds'][mask]
            ev = self.eigenvectors['all_manifolds'][:,mask]
        except KeyError:
            e = self.eigenvalues[manifold_key][mask]
            ev = self.eigenvectors[manifold_key][:,mask]

        psi = psi_obj(t) * np.exp(-1j * e[:,np.newaxis]*t[np.newaxis,:])
        if original_H_basis:
            psi = ev.dot(psi)
        else:
            pass

        self.get_psi_time += time.time() - t_get_psi0

        return psi
        



import numpy as np
import numpy.polynomial.chebyshev as npch
from scipy.interpolate import interp1d as sinterp1d

class perturbative_container:
    """This class is used for storing wavefunctions and density matrices
        as 2D arrays, with the second index being time. They are stored in
        the interaction picture, and are typically assumed to be 0 for 
        time values before those specified, and constant for time values
        after those specified.
    """
    def __init__(self,t,f,bool_mask,pulse_number,manifold_key,pdc,t0,*,
                 interp_kind='linear',interp_left_fill=0,simultaneous=1):
        """f can be either a wavefunction or density matrix, both given as
            2D arrays, with the first index the eigen-index, and the second 
            index being the time
        Args:
            t (np.ndarrray) : 1D array of time values
            f (np.ndarrray) : 2D array representing psi or rho
            bool_mask (np.ndarray) : boolean mask of length of the eigenvalues,
                representing which eigenvectors have non-zero amplitude
            pulse_number (int) : most recent pulse interaction
            manifold_key (str) : which manifold does this psi or rho exist in
            pdc (tuple) : partial phase-discrimination condition
            t0 (float) : interaction picture time-zero value
        Keyword Args:
            interp_kind (str) : type of interpolation to use (e.g. linear, cubic, etc.)
            interp_left_fill (float) : value to use for extrapolation to earlier times
            simultaneous (int) : number of simultaneous pulse interactions (only 
                relevant for impulsive calculations)
"""
        self.bool_mask = bool_mask
        self.pulse_number = pulse_number
        self.manifold_key = manifold_key
        self.pdc = pdc
        self.t0 = t0
        self.pdc_tuple = tuple(tuple(pdc[i,:]) for i in range(pdc.shape[0]))
        self.simultaneous = simultaneous
        
        if t.size == 1:
            if simultaneous < 1:
                raise Exception('keyword argument simultaneous must be an integer greater than 0')
            self.impulsive = True
            n, M = f.shape
            self.M = M+2
            self.n = n
            self.t = np.array([-1,0,1],dtype='float') * np.spacing(t[0]) + t[0]
            f_new = np.zeros((n,3),dtype='complex')
            f_new[:,0] = interp_left_fill
            f_new[:,1] = (1 + interp_left_fill)/2 * f[:,0]/simultaneous
            f_new[:,2] = f[:,0]*2**(simultaneous-1)/simultaneous
            self.asymptote = f_new[:,-1]
            self.f = f_new
            
            self._f = f_new

            self.f_interp = self.impulsive_fun(self.asymptote,left_fill = interp_left_fill)
            
        else:
            self.impulsive = False
            self.t = t
            self.f = f
            self._f = self.extend(f,left_fill = interp_left_fill)
            self.f_interp = self.make_interpolant(kind=interp_kind,
                                                 left_fill=interp_left_fill)
        
    def extend(self,f,*,left_fill = 0):
        """Takes the input psi or rho and creates an array that is three times
            bigger, with constant values padding before and after (the fill 
            values for before are specified, those for after are always a 
            constant extrapolation of the final value of f)

        Args:
            f (np.ndarray): 2D array representing psi or rho
        Keyword Args:
            left_fill (float) : constant extrapolation for times before
                the pulse interacted (usually 0)
        """
        n, M = f.shape
        self.M = M
        self.n = n
        new_f = np.zeros((n,3*M),dtype='complex')
        new_f[:,0:M] = left_fill
        new_f[:,M:2*M] = f
        asymptote = f[:,-1]
        self.asymptote = asymptote
        new_f[:,2*M:] = asymptote[:,np.newaxis]

        return new_f

    def make_interpolant(self,*, kind='cubic', left_fill=0):
        """Interpolates density matrix and pads using left_fill to the 
            left and f[-1] to the right
"""
        left_fill = np.ones(self.n,dtype='complex')*left_fill
        right_fill = self.f[:,-1]
        return sinterp1d(self.t,self.f,fill_value = (left_fill,right_fill),
                         assume_sorted=True,bounds_error=False,kind=kind)

    def impulsive_fun(self,asymptote,left_fill=0):
        if left_fill == 0:
            def f(t):
                zero_val = 0.5**self.simultaneous
                heavi = np.heaviside(t-self.t[1],zero_val)[np.newaxis,:]
                return asymptote[:,np.newaxis] * heavi
        else:
            def f(t):
                try:
                    return asymptote[:,np.newaxis] * np.ones(len(t))[np.newaxis,:]
                except:
                    return asymptote[:,np.newaxis]
        return f

    def __call__(self,t):
        try:
            length = len(t)
            if length == 0:
                return np.array([])
        except TypeError:
            #Assume input must be a number (not a list or an array)
            length = 1
        
        if self.impulsive:
            if length == 1:
                if np.isclose(t,self.t[1],atol=1E-15):
                    return self.f[:,1:2]
                else:
                    return self.f_interp(t)
            else:
                return self.f_interp(t)

        # the following logic is designed to speed up calculations outside of the impulsive limit
        if type(t) is np.ndarray:
            if t.size == 0:
                return np.array([])
            elif t[0] > self.t[-1]:
                if t.size <= self.M:
                    ans = self._f[:,-t.size:]#.copy()
                else:
                    ans = np.ones(t.size,dtype='complex')[np.newaxis,:] * self.asymptote[:,np.newaxis]
            elif t[-1] < self.t[0]:
                if t.size <= self.M:
                    ans = self._f[:,:t.size]#.copy()
                else:
                    ans = np.zeros((self.n,t.size),dtype='complex')
            elif t.size == self.M:
                if np.allclose(t,self.t):
                    ans = self.f#.copy()
                else:
                    ans = self.f_interp(t)
            else:
                ans = self.f_interp(t)
        else:
                ans = self.f_interp(t)
        return ans

    def __getitem__(self,inds):
        return self._f[:,inds]

class ChebPoly:
    def __init__(self,t,f,dom = (-1,1),interp_left_fill = 0,
                 interp_right_fill = 0):
        self.t = t
        self.f = f
        self.dom = dom
        self.interp_left_fill = interp_left_fill
        self.asymptote = np.ones(f.shape[0],dtype='complex') * interp_right_fill
        self.chebpts_to_chebcoef()
        
    def chebpts_to_chebcoef(self):
        """taken from numpy: https://github.com/numpy/numpy/blob/v1.24.0/numpy/polynomial/chebyshev.py#L1780-L1844
        Args:
            yvalues (np.ndarray) : evaluated at chebpts1(order), where 
                order = degree + 1
"""
        self.order = self.t.size
        self.deg = self.order - 1
        self.midpoint = (self.dom[0] + self.dom[1])/2
        self.halfwidth = (self.dom[1] - self.dom[0])/2
        x = (self.t - self.midpoint)/self.halfwidth
        m = npch.chebvander(x, self.deg)
        c = np.dot(m.T, self.f.T)
        c[0] /= self.order
        c[1:] /= 0.5*self.order

        self.cheb_coefs = c.T

        return None

    def f_interp(self,t):
        """Assumes t is sorted
"""
        a,b = self.dom
        if t[0] >= a:
            lowest_ind = 0
        else:
            lowest_ind = np.argmin(np.abs(t-a))
            if t[lowest_ind] < a:
                lowest_ind += 1
        if t[-1] <= b:
            highest_ind = t.size
        else:
            highest_ind = np.argmin(np.abs(t-b))
            if t[highest_ind] < b:
                highest_ind += 1

        ans = np.zeros((self.f.shape[0],t.size),dtype='complex')

        x = (t[lowest_ind:highest_ind] - self.midpoint)/self.halfwidth
        chv = npch.chebvander(x,self.deg)
        ans_interp = chv.dot(self.cheb_coefs.T)
        ans[:,lowest_ind:highest_ind] = ans_interp.T

        if lowest_ind > 0:
            ans[:,:lowest_ind] = self.interp_left_fill

        if highest_ind < t.size:
            high_f = np.ones((t.size - highest_ind),dtype='complex')
            high_f = high_f[np.newaxis,:] * self.asymptote[:,np.newaxis]
            ans[:,highest_ind:] = high_f
            
        return ans

    def integrate(self):
        self.cheb_coefs = npch.chebint(self.cheb_coefs,axis=1,
                                       lbnd=-1)*self.halfwidth
        self.order = self.cheb_coefs.shape[1]
        self.deg = self.order - 1

        return None

    def __call__(self,t):
        return self.f_interp(t)

class cheb_perturbative_container(ChebPoly):
    def __init__(self,t,f,bool_mask,pulse_number,manifold_key,pdc,t0,*,
                 interp_kind='chebyshev',interp_left_fill=0,simultaneous=1,
                 dom = (-1,1)):
        """f can be either a wavefunction or density matrix, both given as
            2D arrays, with the first index the eigen-index, and the second 
            index being the time. Argument interp_kind is ignored
"""
        self.bool_mask = bool_mask
        self.pulse_number = pulse_number
        self.manifold_key = manifold_key
        self.pdc = pdc
        self.t0 = t0
        self.pdc_tuple = tuple(tuple(pdc[i,:]) for i in range(pdc.shape[0]))
        self.simultaneous = simultaneous
        self.dom = dom
        self.interp_left_fill = interp_left_fill
        
        if t.size == 1:
            if simultaneous < 1:
                raise Exception('keyword argument simultaneous must be an integer greater than 0')
            self.impulsive = True
            n, M = f.shape
            self.M = M+2
            self.n = n
            self.t = np.array([-1,0,1],dtype='float') * np.spacing(t[0]) + t[0]
            f_new = np.zeros((n,3),dtype='complex')
            f_new[:,0] = interp_left_fill
            f_new[:,1] = (1 + interp_left_fill)/2 * f[:,0]/simultaneous
            f_new[:,2] = f[:,0]*2**(simultaneous-1)/simultaneous
            self.asymptote = f_new[:,-1]
            self.f = f_new
            
            self._f = f_new

            self.f_interp = self.impulsive_fun(self.asymptote,left_fill = interp_left_fill)
            
        else:
            self.impulsive = False
            self.t = t
            self.f = f
            self._f = self.extend(f,left_fill = interp_left_fill)
            self.chebpts_to_chebcoef()

    def extend(self,f,*,left_fill = 0):
        n, M = f.shape
        self.M = M
        self.n = n
        new_f = np.zeros((n,3*M),dtype='complex')
        new_f[:,0:M] = left_fill
        new_f[:,M:2*M] = f
        asymptote = f[:,-1]
        self.asymptote = asymptote
        new_f[:,2*M:] = asymptote[:,np.newaxis]

        return new_f

    def impulsive_fun(self,asymptote,left_fill=0):
        if left_fill == 0:
            def f(t):
                zero_val = 0.5**self.simultaneous
                heavi = np.heaviside(t-self.t[1],zero_val)[np.newaxis,:]
                return asymptote[:,np.newaxis] * heavi
        else:
            def f(t):
                try:
                    return asymptote[:,np.newaxis] * np.ones(len(t))[np.newaxis,:]
                except:
                    return asymptote[:,np.newaxis]
        return f

    def __call__(self,t):
        try:
            length = len(t)
            if length == 0:
                return np.array([])
        except TypeError:
            #Assume input must be a number (not a list or an array)
            length = 1
        
        if self.impulsive:
            if length == 1:
                if np.isclose(t,self.t[1],atol=1E-15):
                    return self.f[:,1:2]
                else:
                    return self.f_interp(t)
            else:
                return self.f_interp(t)

        # the following logic is designed to speed up calculations outside of the impulsive limit
        if type(t) is np.ndarray:
            if t.size == 0:
                return np.array([])
            elif t[0] > self.t[-1]:
                if t.size <= self.M:
                    ans = self._f[:,-t.size:]#.copy()
                else:
                    ans = np.ones(t.size,dtype='complex')[np.newaxis,:] * self.asymptote[:,np.newaxis]
            elif t[-1] < self.t[0]:
                if t.size <= self.M:
                    ans = self._f[:,:t.size]#.copy()
                else:
                    ans = np.zeros((self.n,t.size),dtype='complex')
            elif t.size == self.M:
                if np.allclose(t,self.t):
                    ans = self.f#.copy()
                else:
                    ans = self.f_interp(t)
            else:
                ans = self.f_interp(t)
        else:
                ans = self.f_interp(t)
        return ans

class RK_perturbative_container:
    def __init__(self,t,f,pulse_number,manifold_key,pdc,*,
                 interp_kind='linear',interp_left_fill=0,simultaneous=1):
        self.pulse_number = pulse_number
        self.manifold_key = manifold_key
        self.pdc = pdc
        self.pdc_tuple = tuple(tuple(pdc[i,:]) for i in range(pdc.shape[0]))
        self.simultaneous = simultaneous

        self.n, self.M = f.shape
        if t.size == 1:
            if simultaneous < 1:
                raise Exception('keyword argument simultaneous must be an integer greater than 0')
            self.impulsive = True
            self.M = self.M+2
            self.t = np.array([-1,0,1],dtype='float') * np.spacing(t[0]) + t[0]
            f_new = np.zeros((self.n,3),dtype='complex')
            f_new[:,0] = interp_left_fill
            f_new[:,1] = (1 + interp_left_fill)/2 * f[:,0]/simultaneous
            f_new[:,2] = f[:,0]*2**(simultaneous-1)/simultaneous
            self.f = f_new
            
            self.interp = self.make_interpolant(kind='zero',left_fill=interp_left_fill)
            
        else:
            self.impulsive = False
            self.t = t
            self.f = f

            self.interp = self.make_interpolant(kind=interp_kind,left_fill=interp_left_fill)

        self.t_checkpoint = self.t
        self.f_checkpoint = self.f

    def make_interpolant(self,*, kind='cubic', left_fill=0):
        """Interpolates density matrix and pads using 0 to the left
            and f[-1] to the right
"""
        left_fill = np.ones(self.n)*left_fill
        right_fill = np.ones(self.n)*np.nan
        fill_value = (left_fill,right_fill)
        return sinterp1d(self.t,self.f, kind=kind,
                         fill_value = fill_value,
                         assume_sorted=True,bounds_error=False)

    def one_time_step(self,f0,t0,tf,*,find_best_starting_time = True):
        if find_best_starting_time and tf < self.t_checkpoint[-1]:
            diff1 = tf - t0

            diff2 = tf - self.t[-1]

            closest_t_checkpoint_ind = np.argmin(np.abs(self.t_checkpoint - tf))
            closest_t_checkpoint = self.t_checkpoint[closest_t_checkpoint_ind]
            diff3 = tf - closest_t_checkpoint

            f0s = [f0,self.f[:,-1],self.f_checkpoint[:,closest_t_checkpoint_ind]]
            
            neighbor_ind = closest_t_checkpoint_ind - 1
            if neighbor_ind >= 0:
                neighbor = self.t_checkpoint[closest_t_checkpoint_ind-1]
                diff4 = tf - neighbor
                f0s.append(self.f_checkpoint[:,neighbor_ind])
            else:
                neighbor = np.nan
                diff4 = np.inf
                

            t0s = np.array([t0,self.t[-1],closest_t_checkpoint,neighbor])
            diffs = np.array([diff1,diff2,diff3,diff4])
            
            for i in range(diffs.size):
                if diffs[i] < 0:
                    diffs[i] = np.inf
            
            if np.allclose(diffs,np.inf):
                raise ValueError('Method extend is only valid for times after the pulse has ended')
            
            t0 = t0s[np.argmin(diffs)]
            f0 = f0s[np.argmin(diffs)]
            
        elif find_best_starting_time and tf > self.t_checkpoint[-1]:
            if self.t_checkpoint[-1] > t0:
                t0 = self.t_checkpoint[-1]
                f0 = self.f_checkpoint[:,-1]
            else:
                pass
            
        else:
            pass

        return self.one_time_step_function(f0,t0,tf,manifold_key=self.manifold_key)

    def extend(self,t):
        ans = np.zeros((self.n,t.size),dtype='complex')
        
        if t[0] >= self.t_checkpoint[0]:

            t_intersect, t_inds, t_checkpoint_inds = np.intersect1d(t,self.t_checkpoint,return_indices=True)

            ans[:,t_inds] = self.f_checkpoint[:,t_checkpoint_inds]

            if t_inds.size == t.size:
                return ans
            else:
                all_t_inds = np.arange(t.size)
                other_t_inds = np.setdiff1d(all_t_inds,t_inds)
                t0 = self.t_checkpoint[-1]
                f0 = self.f_checkpoint[:,-1]
                if t[other_t_inds[0]] >= t0:
                    find_best_starting_time = False
                else:
                    find_best_starting_time = True
                for t_ind in other_t_inds:
                    tf = t[t_ind]
                    ans[:,t_ind] = self.one_time_step(f0,t0,tf,find_best_starting_time = find_best_starting_time)
                    t0 = tf
                    f0 = ans[:,t_ind]
            
        elif t[0] >= self.t[-1]:
            t0 = self.t[-1]
            f0 = self.f[:,-1]
            for i in range(len(t)):
                ans[:,i] = self.one_time_step(f0,t0,t[i],find_best_starting_time = True)
                t0 = t[i]
                f0 = ans[:,i]
        else:
            raise ValueError('Method extend is only valid for times after the pulse has ended')

        self.f_checkpoint = ans
        self.t_checkpoint = t
        return ans

    def __call__(self,t):
        """Assumes t is sorted """
        try:
            length = len(t)
            if length == 0:
                return np.array([])
        except TypeError:
            #Assume input must be a number (not a list or an array)
            length = 1
        if type(t) is np.ndarray:
            pass
        elif type(t) is list:
            t = np.array(t)
        else:
            t = np.array([t])
        extend_inds = np.where(t>self.t[-1])
        interp_inds = np.where(t<=self.t[-1])
        ta = t[interp_inds]
        tb = t[extend_inds]
        if ta.size > 0:
            ans_a_flag = True
            if ta.size == self.M and np.allclose(ta,self.t):
                ans_a = self.f
            else:
                ans_a = self.interp(ta)
        else:
            ans_a_flag = False
        if tb.size > 0:
            ans_b = self.extend(tb)
            ans_b_flag = True
        else:
            ans_b_flag = False
            
        if ans_a_flag and ans_b_flag:
            ans = np.hstack((ans_a,ans_b))
        elif ans_a_flag:
            ans = ans_a
        elif ans_b_flag:
            ans = ans_b
        else:
            ans = None
        return ans

    def __getitem__(self,inds):
        return self.f[:,inds]
